import logging
import numpy as np
import matplotlib.pylab as plt

logger = logging.getLogger('neutReac.los')
logger.setLevel(logging.DEBUG)


class CELL:
    pass


class DETECTOR:
    pass


class PHILOS:

    def __init__(self, par_d, npoints=5000):

        theta  = par_d['theta']
        phi_in = par_d['phi'] 
        x0     = par_d['x0']
        y0     = par_d['y0']
        z0     = par_d['z0']
        xend   = par_d['xend']
        phi0 = np.arctan2(x0, y0)
        phi  = phi_in + phi0
        dx   = xend - x0
        dr   = dx/np.sin(phi)
        self.dr = np.abs(dr)
        yend = y0 + dr*np.cos(phi)
        zend = z0 + self.dr*np.tan(theta)
        self.xline = np.linspace(x0, xend, npoints)
        self.yline = np.linspace(y0, yend, npoints)
        self.zline = np.linspace(z0, zend, npoints)
        self.rline = np.hypot(self.xline, self.yline)


class XYLOS:

    def __init__(self, par_d, npoints=5000):

        theta  = par_d['theta']
        x0 = par_d['x0']
        y0 = par_d['y0']
        z0 = par_d['z0']
        xend = par_d['xend']
        yend = par_d['yend']
        dx = xend - x0
        dy = yend - y0
        self.dr = np.hypot(dx, dy)
        zend = z0 + self.dr*np.tan(theta)
        self.xline = np.linspace(x0, xend, npoints)
        self.yline = np.linspace(y0, yend, npoints)
        self.zline = np.linspace(z0, zend, npoints)
        self.rline = np.hypot(self.xline, self.yline)


class DETECTOR_LOS:

    def __init__(self, geo):
        self.geo = geo
        self.run()

    def run(self):

        los_d = {}
        los_d['x0']    = -self.geo['y_det']
        los_d['y0']    = 0.001
        los_d['z0']    = self.geo['z_det']
        los_d['xend']  = 0 # 1 plasma crossing
#        los_d['xend']  = self.geo['y_det'] # 2 plasma crossings
        los_d['theta'] = np.radians(self.geo['tilt'])
        los_d['phi']   = -np.arctan2(self.geo['tan_radius'], self.geo['y_det'])

        ctilt = np.cos(los_d['theta'])
        dy = self.geo['disk_thick']*ctilt
        ndisks = int(-2*self.geo['y_det']/dy)

        self.det = DETECTOR()
        self.det.los = PHILOS(los_d, npoints=ndisks)

# Get separatrix {R,z}

        rmin = self.geo['Rmaj']  - self.geo['r_chamb'] 
        rmax = self.geo['Rmaj']  + self.geo['r_chamb'] 

# Restrict to LOS inside the plasma

        dl = np.hypot(self.det.los.rline[1] - self.det.los.rline[0], self.det.los.zline[1] - self.det.los.zline[0])
        ind = (self.det.los.rline > rmin - dl) & (self.det.los.rline < rmax + dl)
        r_in  = self.det.los.rline[ind]
        self.det.los.z = self.det.los.zline[ind]
        self.det.los.y = self.det.los.xline[ind]
        y_in  = self.det.los.yline[ind]

# Write ASCII file for LOS, used in the spectrum evaluation

        self.det.pos = (self.geo['tan_radius'], self.geo['y_det'], self.geo['z_det'])
        det_dist = np.hypot(self.det.los.y - self.det.pos[1], self.det.los.z - self.det.pos[2])
        ctilt = np.cos(np.radians(self.geo['tilt']))
        stilt = np.sin(np.radians(self.geo['tilt']))
        dy = self.geo['disk_thick']*ctilt
        coll_rad = 0.5*self.geo['coll_diam']
        dist_corr = self.geo['d_det_coll']/(1. + self.geo['det_radius']/coll_rad) # Shifting the cone vertex to the point where lines (det-left - coll-right, det-right - coll-left) cross
        self.det.tan_cone_aper = coll_rad/dist_corr
#        logger.debug(coll_rad, dist_corr, tan_cone_aper)
        offset = self.geo['d_det_coll'] - dist_corr # scalar, r2
        disk_radius = (det_dist - offset)*self.det.tan_cone_aper # array on discretised LoS
        n_disks = len(self.det.los.y)

# Each disk is divided in n_circles circular sectors,
# n_circles depends on the disk radius (which is almost constant)
# Every circular sector is divided in n_sectors sectors,
# equidistant poloidally; n_sectors is proportional to the radius
# of the circular sector

        self.cell = CELL()
        self.cell.x     = []
        self.cell.y     = []
        self.cell.z     = []
        self.cell.omega = []
        self.cell.vol   = []

        for jdisk in range(n_disks):
            n_circles = int(0.5 + disk_radius[jdisk]/self.geo['cell_radius'])
            delta_radius = disk_radius[jdisk]/float(n_circles)

# radius, alpha in the 'middle' of the sector
# The central circle has only one sector (cell)
            cvol = np.pi * delta_radius**2 * dy
            self.cell.vol = np.append(self.cell.vol, np.repeat(cvol, n_circles**2))

            radius = (0.5 + np.arange(n_circles))*delta_radius
            radius[0] = 0.
            omega_fac = np.pi * self.geo['det_radius']**2/det_dist[jdisk]**3
            cell_detDist = np.hypot(det_dist[jdisk], radius)
            omegaCircle = omega_fac * cell_detDist
            for j_circle in range(n_circles):
                n_sectors = 2*j_circle + 1
                self.cell.omega += n_sectors*[omegaCircle[j_circle]]
# Poloidal sectors (cells) in a circle
                alpha = np.linspace(0, 2.*np.pi, n_sectors, endpoint=False)
                rcos  = list(radius[j_circle]*np.cos(alpha))
                rssin = list(-self.det.los.y[jdisk] - radius[j_circle]*np.sin(alpha)*stilt)
                rcsin = list( self.det.los.z[jdisk] + radius[j_circle]*np.sin(alpha)*ctilt)
# cell_pos: with respect to torus center
                self.cell.x += rcos
                self.cell.y += rssin
                self.cell.z += rcsin
        self.cell.x = np.array(self.cell.x) + self.det.pos[0]
        self.cell.y = np.array(self.cell.y)
        self.cell.z = np.array(self.cell.z)
        self.cell.omega = np.array(self.cell.omega)
        logger.info('Done LOS cone calculation')


    def writeLOS(self):

        n_cells = len(self.cell.vol)
        n_disks = len(self.det.los.y)
        cone_aper = np.degrees(np.arctan(self.det.tan_cone_aper))
        ctilt = np.cos(np.radians(self.geo['tilt']))
        dy = self.geo['disk_thick']*ctilt

        header = \
           'LOS\n' + \
           '   y1 = %9.4f m\n'  % self.det.los.y[0]  + \
           '   y2 = %9.4f m\n'  % self.det.los.y[-1] + \
           '   z1 = %9.4f m\n'  % self.det.los.z[0]  + \
           '   z2 = %9.4f m\n'  % self.det.los.z[-1] + \
           'Detector:\n' + \
           '   Position [m] x, y, z: %9.4f, %9.4f, %9.4f\n' %self.det.pos + \
           '   Radius = %9.4f m\n' % self.geo['det_radius'] + \
           '   Collimation angle = %9.4f deg\n' % cone_aper + \
           'Disks:\n' + \
           '   Thickness = %9.4f m\n' % dy + \
           '   # disks = %5d\n' % n_disks + \
           'Cells:\n' + \
           '   Radius = %9.4f m\n' % self.geo['cell_radius'] + \
           '   # cells = %d\n' %n_cells + \
"""
   (x,y,x) cell cartensian coordinates [m]
   Omega  Steradians is the volume in the solid angle
   Vol = cell volume [m**3]
x             y             z             Omega         Vol
"""

        los_file = 'los/%s.los' %self.geo['label']

        logger.info('Storing ASCII output, n_cells=%d', n_cells)
        np.savetxt(los_file, np.hstack((self.cell.x, self.cell.y, self.cell.z, self.cell.omega, self.cell.vol)).reshape(5, n_cells).T, header=header, fmt='%13.6E')
        logger.info('Written output file %s' %los_file)


    def plotLOS(self, show=False, figsize=(8.8, 5.9)):

        R_m = np.hypot(self.cell.x, self.cell.y)

        fig = plt.figure(1, figsize=figsize)
        fig.subplots_adjust(left=0.05, bottom=0.1, right=0.95, top=0.98, wspace=0.2)

        ax1 = plt.subplot(1, 2, 1, aspect='equal')
        ax1.set_xlim([0.5, 3.5])
        ax1.set_ylim([-1.5, 1.5])
        ax1.set_xlabel('R [m]', labelpad=2)
        ax1.set_ylabel('z [m]', labelpad=2)
        ax1.plot(R_m, self.cell.z, 'ro') 

        ax2 = plt.subplot(1, 2, 2, aspect='equal')
        ax2.set_xlim([-3, 3])

# Plot AUG wall
        try:
            import aug_sfutils as sf
        except:
            logger.error('Missing vessel contour data')
        if 'sf' in locals():
            gc_d  = sf.getgc()
            tor_d = sf.getgc_tor() 
            for gc in gc_d.values():
                ax1.plot(gc.r, gc.z, 'b-')
            for tor in tor_d.values():
                ax2.plot(tor.x, tor.y, 'b-')

        ax2.plot(self.cell.x, self.cell.y, 'ro')
        ax2.set_xlabel('x [m]', labelpad=2)
        ax2.set_ylabel('y [m]', labelpad=2)
        ax2.tick_params(which='major', length=4, width=0.5)

        if show:
            plt.show()
        else:
            return fig
        



import logging
import numpy as np
import h5py
from scipy.io import netcdf_file
from scipy.interpolate import RectBivariateSpline

logger = logging.getLogger('DRESS LoS.fi')
logger.setLevel(logging.DEBUG)

flt = np.float64


class TRANSP2DRESS:
    '''Reads TRANSP CDF and NUBEAM cdf output files, converting units into DRESS'''

    def __init__(self, f_plasma, f_fast):

        logger.info('Reading TRANSP distribution')

        cv = netcdf_file(f_plasma, 'r', mmap=False).variables
        fv = netcdf_file(f_fast, 'r', mmap=False).variables

        spc_lbl = b''.join(fv['SPECIES_1'].data).decode()
        tfbm = fv['TIME' ].data
        tim  = cv['TIME3'].data
        jt = np.argmin(np.abs(tim - tfbm))

# Separatrix
        self.Rbnd = 1e-2*fv['RSURF'].data[-1, :]
        self.Zbnd = 1e-2*fv['ZSURF'].data[-1, :]

        rhot_cdf = cv['X'].data[jt, :]
        rhot_fbm = fv['X2D'].data

        self.code_d = {}
        self.code_d['E']        = 1.e-3*fv['E_%s' %spc_lbl].data # eV -> keV
        self.code_d['pitch']    = fv['A_%s' %spc_lbl].data
        self.code_d['R']        = 1e-2*fv['R2D'].data
        self.code_d['Z']        = 1e-2*fv['Z2D'].data
        self.code_d['rho']      = rhot_fbm
        self.code_d['density']  = fv['bdens2'].data
        self.code_d['Ti']       = 1e-3*np.interp(rhot_fbm, rhot_cdf, cv['TI'].data[jt, :]) # eV -> keV
        self.code_d['nd']       =  1e6*np.interp(rhot_fbm, rhot_cdf, cv['ND'].data[jt, :]) # 1/cm**3 -> 1/m**3
        self.code_d['ang_freq'] =      np.interp(rhot_fbm, rhot_cdf, cv['OMEGA'].data[jt, :])
        self.code_d['dV']       = 1e-6*fv['BMVOL'].data # cm**3 -> m**3
        self.code_d['F']        = np.transpose(fv['F_%s' %spc_lbl].data, axes=(0, 2, 1)) # -> cell, E, pitch; no need to normalise


class ASCOT2DRESS:
    '''Reads ASCOT HDF5 output file, converting units into DRESS'''


    def __init__(self, f_fast):

        logger.info('Reading ASCOT distribution')

# Reading HDF5 file

        unit_d = {}
        f = h5py.File(f_fast, 'r')

        pl1d       = f['plasma/1d']
        bfield     = f['bfield']
        dist_grids = f['distributions/rzPitchEdist/abscissae']
        fbm_h5     = f['distributions/rzPitchEdist/ordinate'][:] # species, time, energy, pitch, z, R, ?

        rhop_pl  = pl1d['rho'][:]
        ti = 1.e-3*pl1d['ti' ][:]    # eV -> keV
        nD =       pl1d['ni' ][:, 0] # m**-3
        vt =       pl1d['vtor'][:]   # m/s
        for lbl in ('ti', 'ni', 'vtor'):
            unit_d[lbl] = pl1d[lbl].attrs['unit'].decode()

        (Raxis, ) = bfield['raxis']
        (zaxis, ) = bfield['zaxis']
        R_psi = bfield['r'][:]
        z_psi = bfield['z'][:]
        psi2d = bfield['2d/psi'][:].T # R, z

        grid = {}
        grid_b = {}
        for var in dist_grids.values():
            lbl  = var.attrs['name'].decode()
            unit = var.attrs['unit'].decode()
            unit_d[lbl] = unit
            grid_b[lbl] = var[:]
            grid[lbl] = 0.5*(var[1:] + var[:-1])

        f.close()

# R, z rectangular domain
        Rmin = grid_b['R'][0]
        Rmax = grid_b['R'][-1]
        Zmin = grid_b['z'][0]
        Zmax = grid_b['z'][-1]
        self.Rbnd = [Rmin, Rmax, Rmax, Rmin]
        self.Zbnd = [Zmin, Zmin, Zmax, Zmax]

# Fast ion distribution and grids
        fbm = fbm_h5[0, 0, :, ::-1, :, :, 0] # energy, pitch, z, R; no need to normalise
        nE, nmu, nz, nR = fbm.shape
        nRz = nR*nz
        dR  = (Rmax - Rmin)/float(nR)
        dz  = (Zmax - Zmin)/float(nz)
        dE  = (grid_b['energy'][-1] - grid_b['energy'][0])/float(nE)
        dmu = (grid_b['pitch' ][-1] - grid_b['pitch' ][0])/float(nmu)

        self.code_d = {}
        Rmesh, Zmesh = np.meshgrid(grid['R'], grid['z'], indexing='ij') # 2D grid
        self.code_d['R'] = Rmesh.ravel() # Keep everything 1d, 1st index is R
        self.code_d['Z'] = Zmesh.ravel()
        self.code_d['E'] = grid['energy']*6.242e+15 # J -> keV
        self.code_d['pitch'] = grid['pitch']
        self.code_d['dV'] = 2.*np.pi*self.code_d['R']*dR*dz
        self.code_d['F'] = np.transpose(fbm, axes=(3, 2, 0, 1)).reshape(nRz, nE, nmu)  # Rz, energy, pitch

# 2D grid for rho_pol
        psi_sep = 0
        f_psi = RectBivariateSpline(R_psi, z_psi, psi2d)
        psi_axis = f_psi(Raxis, zaxis)[0][0]
        psi_red  = f_psi(self.code_d['R'], self.code_d['Z'], grid=False)
        psi_norm = (psi_red - psi_axis)/(psi_sep - psi_axis)
        self.code_d['rho'] = np.sqrt(psi_norm)

# Integrate distribution over E, pitch
        self.code_d['density'] = np.sum(self.code_d['F'], axis=(1, 2))*dE*dmu

# Interpolate Ti, nd, v_rot to (unrolled) 2D grid
        self.code_d['Ti']    = np.zeros(nRz, dtype=flt)
        self.code_d['nd']    = np.zeros(nRz, dtype=flt)
        self.code_d['v_rot'] = np.zeros((nRz, 3), dtype=flt)
        nrho_pl = len(rhop_pl)
        for jR in range(nR):
            for jz in range(nz):
                xrho = self.code_d['rho'][jR + jz*nR]
                if np.isnan(xrho) or xrho > 1.: # outside separatrix
                    self.code_d['Ti'][jR + jz*nR] = np.nan
                    self.code_d['nd'][jR + jz*nR] = np.nan
                else:
                    jrho = np.argmin(np.abs(rhop_pl - xrho))
                    if jrho == 0:
                        ind = [jrho, jrho + 1]
                    elif jrho == nrho_pl-1:
                        ind = [jrho-1, jrho]
                    else:
                        ind  = [jrho-1, jrho, jrho+1]
                    self.code_d['Ti'   ][jR + jz*nR] = np.interp(xrho, rhop_pl[ind], ti[ind])
                    self.code_d['nd'   ][jR + jz*nR] = np.interp(xrho, rhop_pl[ind], nD[ind])
#                    self.code_d['v_rot'][jR, jz, 1] = np.interp(xrho, rhop_pl[ind], vt[ind]) # Too high in repo case



import numpy as np
from .reactions import reaction
from .constants import alpha

# ControlRoom/source/reactivities.cpp
#
# neutrons/paper/bosch_nf92.pdf
# Bosch-Hale, table VII
# Ti validity range (keV)
#   DT    : 0.2-100
#   D3He  : 0.5-190
#   DDn3He: 0.2-100
#   DDpt  : 0.2-100


def react(T_keV, reac_lbl):
    '''Input:
        Ti: reactant temperature (scalar or 1d array) in keV
    '''

# Reduced mass, MeV -> keV

    reac = reaction[reac_lbl]
    coeff = reac.coeff_reac
    mu_c2 = 1e3*(reac.in1.m * reac.in2.m)/(reac.in1.m + reac.in2.m)

    Bg = np.pi * alpha * reac.in1.Z * reac.in2.Z * np.sqrt(2 * mu_c2)
    theta = T_keV/(  1 - \
        (T_keV * (coeff[1] + T_keV * (coeff[3] + T_keV * coeff[5]))) / \
        (1 + T_keV * (coeff[2] + T_keV * (coeff[4] + T_keV * coeff[6])))  \
    )
    csi = (Bg**2/(4*theta))**0.333
    react = coeff[0] * theta * np.sqrt(csi/(mu_c2 * T_keV**3)) * np.exp(-3*csi)

    return react



import scipy.constants as sc

con = sc.physical_constants
amu = con['atomic mass constant energy equivalent in MeV'][0]
mpc2 = con['proton mass energy equivalent in MeV'][0]
mDc2 = con['deuteron mass energy equivalent in MeV'][0]
mTc2 = con['triton mass energy equivalent in MeV'][0]
mnc2 = con['neutron mass energy equivalent in MeV'][0]
mHe3c2 = 3.014932150*amu # MeV
mac2 = con['alpha particle mass energy equivalent in MeV'][0]
mHe4c2 = con['alpha particle mass energy equivalent in MeV'][0]
alpha = sc.alpha # Fine structure constant
epsilon0 = sc.epsilon_0
echarge = sc.elementary_charge
c = sc.c



import os, json, logging
import numpy as np
from scipy.interpolate import interp1d, RectBivariateSpline
import matplotlib.pylab as plt

logger = logging.getLogger('nresp.cs')
logger.setLevel(level=logging.DEBUG)

nrespDir = os.path.dirname(os.path.realpath(__file__))
crossDir='%s/cross-sections' %nrespDir
flt_typ = np.float64

class crossSections:

    
    def __init__(self, json=True):
        
        if json:
            self.fromJSON()


    def fromJSON(self):
        '''Read cross-sections from JSON files'''

        logger.info('Reading JSON cross-sections from dir "%s"', crossDir)
        f_json = '%s/crossSections.json' %crossDir
        with open(f_json, 'r') as fjson:
            crSec = json.load(fjson)
        self.EgridTot  = crSec['EgridTot']
        self.Egrid = 0.02*np.arange(1001)
        f_json = '%s/alphas3.json' %crossDir
        with open(f_json, 'r') as fjson:
            self.alphas3 = json.load(fjson)
        CSalphas3 = np.array(self.alphas3['crossSec'], dtype=flt_typ).T

        self.int_alphas3 = interp1d(self.alphas3['Egrid'], CSalphas3, axis=1, assume_sorted=True, kind='linear', fill_value='extrapolate')
        self.max_level = CSalphas3.shape[0]

        self.crSec_d  = {}
        csDiff_d = {}
        self.reacTot  = []
        self.reacDiff = []
        for reac in crSec['reactions']:
            f_json = '%s/%s.json' %(crossDir, reac)
            if os.path.isfile(f_json):
                with open(f_json, 'r') as fjson:
                    self.crSec_d[reac] = json.load(fjson)
                    if 'crossTot' in self.crSec_d[reac].keys():
                        self.reacTot.append(reac)
                        if 'EgridTot' not in self.crSec_d[reac].keys():
                            nE = len(self.crSec_d[reac]['crossTot'])
                            self.crSec_d[reac]['EgridTot'] = self.EgridTot[-nE:]
                    if 'crossDiff' in self.crSec_d[reac].keys():
                        self.reacDiff.append(reac)
                        crSecArray = 1.e-4*np.array(self.crSec_d[reac]['crossDiff'], dtype=flt_typ)
                        n_the, nE = crSecArray.shape
                        if 'n_theta' not in globals():
                            n_theta = n_the + 2
                            theta_grid = np.linspace(0, 1, n_theta, endpoint=True)
                        csDiff_d[reac] = np.vstack((np.zeros(nE), crSecArray, np.pi + np.zeros(nE))).T

        self.reacTotUse = self.reacTot[:10] # exclude CarTot, AlTot, HE1, HE2

        self.csd_d = {}
        for reac in self.reacDiff:
            self.csd_d[reac] = RectBivariateSpline(self.crSec_d[reac]['EgridDiff'], theta_grid, csDiff_d[reac], kx=2, ky=2)

        self.cst1d = {}
        for reac in self.reacTot:
            Interp = interp1d(self.crSec_d[reac]['EgridTot'], self.crSec_d[reac]['crossTot'], kind='linear', assume_sorted=True, fill_value='extrapolate')
            self.cst1d[reac] = Interp(self.Egrid)


    def cosInterpReac2d(self, reac, En_in, randomAngle):

        if reac in self.reacDiff:
            return np.cos(self.csd_d[reac](En_in, randomAngle, grid=False))
        else:
            logger.error('No differential cross-section for label "%s"', reac)
            return None


    def plot(self):

        plt.figure('Cross-sections', (14, 5))
        cs_c  = np.zeros_like(np.float32(self.EgridTot))
        cs_al = np.zeros_like(np.float32(self.EgridTot))
        for reac in self.reacTot:
            if reac[:6] == "12C(N,":
                plt.subplot(1, 3, 2)
                En = self.crSec_d[reac]['EgridTot']
                cs = self.crSec_d[reac]['crossTot']
                cs_tmp = np.interp(self.EgridTot, En, cs)
                cs_c += cs_tmp
                plt.plot(En, cs, label=reac)
            elif reac[:8] == "27AL(N,N":
                plt.subplot(1, 3, 3)
                En = self.crSec_d[reac]['EgridTot']
                cs = self.crSec_d[reac]['crossTot']
                cs_tmp = np.interp(self.EgridTot, En, cs)
                cs_al += cs_tmp
                plt.plot(En, cs, label=reac)

        plt.subplot(1, 3, 1)
        reac = "H(N,N)H"
        plt.plot(self.crSec_d[reac]['EgridTot'], self.crSec_d[reac]['crossTot'], label=reac)
        plt.legend()

        plt.subplot(1, 3, 2)
        plt.ylim([0, 5000])
        reac = 'CarTot'
        plt.plot(self.crSec_d[reac]['EgridTot'], self.crSec_d[reac]['crossTot'], label=reac)
        plt.plot(self.EgridTot, cs_c, label='Sum')
        plt.legend()

        plt.subplot(1, 3, 3)
        plt.ylim([0, 6000])
        reac = 'AlTot'
        plt.plot(self.crSec_d[reac]['EgridTot'], self.crSec_d[reac]['crossTot'], label=reac)
        plt.plot(self.EgridTot, cs_al, label='Sum')
        plt.legend()
    
        plt.show()

#  LocalWords:  interp



#!/usr/bin/env python

"""NEutron REactions Utility Software
"""

import os, logging

__author__  = 'Giovanni Tardini (Tel. +49 89 3299-1898)'
__version__ = '0.0.3'
__date__    = '15.01.2025'

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%Y-%m-%d %H:%M:%S')
hnd = logging.StreamHandler()
hnd.setFormatter(fmt)
hnd.setLevel(level=logging.INFO)
logger = logging.getLogger('nereus')
logger.addHandler(hnd)
logger.setLevel(logging.INFO)
logger.propagate = False

nereus_home = os.path.dirname(os.path.realpath(__file__))

logger.info('Using version %s', __version__)
logger.info('NEREUS home %s', nereus_home)

from .nereus_gui import *



import os, logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches
from multiprocessing import Pool, cpu_count
import dress
import response
from dress_client import fi_codes

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%H:%M:%S')
hnd = logging.StreamHandler()
hnd.setFormatter(fmt)
logger = logging.getLogger('DRESS LoS')
logger.addHandler(hnd)
logger.setLevel(logging.DEBUG)

flt = np.float64


def calcvols(tuple_in):

    vols, dist1, dist2, scalc, En_bins = tuple_in
    nes = dress.utils.calc_vols(vols, dist1, dist2, scalc, En_bins, integrate=True, quiet=True)

    return nes


class nSpectrum:


    def __init__(self, f_in1, f_in2, f_los=None, src='transp', samples_per_volume_element=1e4):

        self.samples_per_volume_element = samples_per_volume_element
        self.code = src

        if src == 'transp':
            self.codeClass = fi_codes.TRANSP2DRESS(f_in1, f_in2)
        elif src == 'ascot':
            self.codeClass = fi_codes.ASCOT2DRESS(f_in1)

        if f_los is None:
            self.los = False
            self.dressInput = self.codeClass.code_d
            self.dressInput['solidAngle'] = 4*np.pi*np.ones_like(self.dressInput['dV'])
        else:
            self.los = True
            x_m, y_m, z_m, omega, V_m3 = np.loadtxt(f_los, unpack=True)
            R_m = np.hypot(x_m, y_m)
            self.los_dressInput(R_m, z_m)
            self.dressInput['R']  = R_m[self.inside]
            self.dressInput['Z']  = z_m[self.inside]
            self.dressInput['dV'] = V_m3[self.inside]
            self.dressInput['solidAngle'] = omega[self.inside]


    def los_dressInput(self, R_m, z_m):
        '''Mapping quantities from original volumes to LoS volumes. Removing LoS volumes outside a {R, z} domain (separatrix or 2D cartesian grid)'''

        n_los = len(R_m)

# LoS volumes inside R, z domain

        RZ_points = np.hstack((R_m, z_m)).reshape(2, n_los).T
        sepPolygon = np.hstack((self.codeClass.Rbnd, self.codeClass.Zbnd)).reshape(2, len(self.codeClass.Rbnd)).T
        self.sepPath = Path(sepPolygon)
        self.inside = self.sepPath.contains_points(RZ_points)
        logger.debug('Volumes inside Sep %d out of %d', np.sum(self.inside), n_los)

        los2fbm = np.zeros(n_los, dtype=np.int32)
        for jlos in range(n_los):
            if self.inside[jlos]:
                d2 = (self.codeClass.code_d['R'] - R_m[jlos])**2 + (self.codeClass.code_d['Z'] - z_m[jlos])**2
                los2fbm[jlos] = np.argmin(d2)
        los_sep = los2fbm[self.inside]

        self.dressInput = {}
        for key, val in self.codeClass.code_d.items():
            if key in ('E', 'pitch'):
                 self.dressInput[key] = val
            else: # map variables onto LoS volumes
                self.dressInput[key] = val[los_sep]


    def run(self, parallel=True):
        '''Execute DRESS calculation, computing Neutron Emission Spectra: beam-target, thermonuclear, beam-beam'''

        logger.info('Running DRESS, #MC %d', self.samples_per_volume_element)

        Ncells = len(self.dressInput['rho'])

        Bdir = np.atleast_2d([0, -1, 0])
        B_dir = np.repeat(Bdir, Ncells, axis=0)   # B-dir for each spatial location

        if 'v_rot' not in self.dressInput.keys():
            self.dressInput['v_rot'] = np.zeros((Ncells, 3), dtype=flt)
            if 'ang_freq' in self.dressInput.keys():
                self.dressInput['v_rot'][:, 1] = self.dressInput['ang_freq']*self.dressInput['R']

        dd    = dress.reactions.DDNHe3Reaction()
        scalc = dress.SpectrumCalculator(dd, n_samples=self.samples_per_volume_element)

# Neutron energy bins [keV]
        bin_keV = 10.
        En_bins = np.arange(1500, 3500, bin_keV)    # bin edges
        self.En = 0.5*(En_bins[1:] + En_bins[:-1])  # bin centers

# Compute spectra components

        if parallel:
# Split arrays

            self.dressSplit = {}
            n_split = 20
            for key in ('dV', 'solidAngle', 'R', 'Z', 'density', 'F', 'nd', 'Ti', 'v_rot'):
                self.dressSplit[key] = np.array_split(self.dressInput[key], n_split)
            vols_spl = {}
            fast_spl = {}
            bulk_spl = {}
            for j in range(n_split):
                Nvols = len(self.dressSplit['dV'][j])
                B_dir = np.repeat(Bdir, Nvols, axis=0)   # B-dir for each spatial location
                vols_spl[j] = dress.utils.make_vols(self.dressSplit['dV'][j], self.dressSplit['solidAngle'][j], pos=(self.dressSplit['R'][j], self.dressSplit['Z'][j]))
                fast_spl[j] = dress.utils.make_dist('energy-pitch', 'd', Nvols, self.dressSplit['density'][j],
                energy_axis=self.dressInput['E'], pitch_axis=self.dressInput['pitch'], distvals=self.dressSplit['F'][j], ref_dir=B_dir)
                bulk_spl[j] = dress.utils.make_dist('maxwellian', 'd', Nvols, self.dressSplit['nd'][j], temperature=self.dressSplit['Ti'][j], v_collective=self.dressSplit['v_rot'][j])

            timeout_pool = 2000
            pool = Pool(cpu_count())
            logger.info('Computing beam-target')
            bt = pool.map_async( calcvols, [(vols_spl[j], fast_spl[j], bulk_spl[j], scalc, En_bins) for j in range(n_split)]).get(timeout_pool)
            logger.info('Computing thermonuclear')
            th = pool.map_async( calcvols, [(vols_spl[j], bulk_spl[j], bulk_spl[j], scalc, En_bins) for j in range(n_split)]).get(timeout_pool)
            logger.info('Computing beam-beam')
            bb = pool.map_async( calcvols, [(vols_spl[j], fast_spl[j], fast_spl[j], scalc, En_bins) for j in range(n_split)]).get(timeout_pool)
            bt = np.array(bt)
            th = np.array(th)
            bb = np.array(bb)
            self.bt = np.sum(bt, axis=0)
            self.th = np.sum(th, axis=0)
            self.bb = np.sum(bb, axis=0)
        else:
            vols = dress.utils.make_vols(self.dressInput['dV'], self.dressInput['solidAngle'], pos=(self.dressInput['R'], self.dressInput['Z']))
            fast_dist = dress.utils.make_dist('energy-pitch', 'd', Ncells, self.dressInput['density'],
                energy_axis=self.dressInput['E'], pitch_axis=self.dressInput['pitch'], distvals=self.dressInput['F'], ref_dir=B_dir)
            bulk_dist = dress.utils.make_dist('maxwellian', 'd', Ncells, self.dressInput['nd'], temperature=self.dressInput['Ti'], v_collective=self.dressInput['v_rot'])
            logger.debug('T #nan: %d, #T<=0: %d', np.sum(np.isnan(bulk_dist.T)), np.sum(bulk_dist.T <= 0))
            logger.debug('nd #nan: %d, #nd<=0: %d', np.sum(np.isnan(bulk_dist.density)), np.sum(bulk_dist.density <= 0))
            logger.info('Computing beam-target')
            self.bt = dress.utils.calc_vols(vols, fast_dist, bulk_dist, scalc, En_bins, integrate=True, quiet=False)
            logger.info('Computing thermonuclear')
            self.th = dress.utils.calc_vols(vols, bulk_dist, bulk_dist, scalc, En_bins, integrate=True, quiet=False)
            logger.info('Computing beam-beam')
            self.bb = dress.utils.calc_vols(vols, fast_dist, fast_dist, scalc, En_bins, integrate=True, quiet=False)     

        for spec in self.bt, self.bb, self.th:
            spec /= bin_keV
        self.bb *= 0.5
        self.th *= 0.5

        rate_bt = bin_keV*np.sum(self.bt)
        rate_th = bin_keV*np.sum(self.th)
        rate_bb = bin_keV*np.sum(self.bb)
        if self.los: # LoS
            logger.info('b-t rate at detector %12.4e N/s', rate_bt)
            logger.info('th  rate at detector %12.4e N/s', rate_th)
            logger.info('b-b rate at detector %12.4e N/s', rate_bb)
            logger.info('Neutron rate at detector %12.4e N/s', rate_bt + rate_th + rate_bb)
        else: # Total
            logger.info('Total b-t neutrons %12.4e N/s', rate_bt)
            logger.info('Total th  neutrons %12.4e N/s', rate_th)
            logger.info('Total b-b neutrons %12.4e N/s', rate_bb)
            logger.info('Total neutron rate %12.4e N/s', rate_bt + rate_th + rate_bb)


    def nes2phs(self, f_resp='responses/rm_bg.cdf'):

        resp = response.RESP()
        fname, ext = os.path.splitext(f_resp)
        if ext.lower() == '.cdf':
            resp.from_cdf(f_resp)
        else:
            resp.from_hepro(f_resp)

        En_MeV = 1e-3*self.En
        nbins = len(resp.Ephs_MeVee)
        self.phs = {}
        self.phs['Elight_MeVee'] = resp.Ephs_MeVee

        for reac in ('bt', 'bb', 'th'):
            self.phs[reac] = np.zeros(nbins, dtype=flt)
            for jEn, En in enumerate(En_MeV):
                dist = (resp.En_MeV - En)**2
                jclose = np.argmin(dist)
                self.phs[reac] += self.__dict__[reac][jEn]*resp.RespMat[jclose, :]


    def storeSpectra(self, f_out='dress_client/output/Spectrum.dat'):
        '''Store ASCII output for DRESS Neutron Emission Spectra'''

        header = 'Eneut      Thermonucl. Beam-target Beam-beam'
        np.savetxt(f_out, np.c_[self.En, self.th, self.bt, self.bb], fmt='%11.4e', header=header)
        logger.info('Stored file %s', f_out)


    def fromFile(self, f_spec):
        '''Read (e.g. for plotting) DRESS ASCII output files, instead of computing Neutron Emission Spectra'''

        self.En, self.th, self.bt, self.bb = np.loadtxt(f_spec, unpack=True, skiprows=1)


    def plotInput(self, jcell=100):
        '''Plot some DRESS input quantities'''

        logger.info('Plotting input')
        plt.figure()
        plt.plot(self.dressInput['rho'], self.dressInput['nd'], 'k+')
        plt.title('Bulk D density')
        plt.xlabel('rho')
        plt.ylabel('density (particles/m^3)')

        plt.figure()
        plt.plot(self.dressInput['rho'], self.dressInput['Ti'], 'go')
        plt.title('Ion temperature')
        plt.xlabel('rho')
        plt.ylabel('temperature (keV)')

        plt.figure()
        plt.plot(self.dressInput['rho'], self.dressInput['v_rot'], 'ro')
        plt.title('Toroidal velocity')
        plt.xlabel('rho')
        plt.ylabel('Toroidal velocity (m/s)')

# Plot fast D density
        plt.figure()
        plt.tripcolor(self.dressInput['R'], self.dressInput['Z'], self.dressInput['density'])
        if hasattr(self, 'sepPath'):
            patch = patches.PathPatch(self.sepPath, facecolor='None', lw=2)
            plt.gca().add_patch(patch)
        plt.title('Fast D density')
        plt.xlabel('R (m)')
        plt.ylabel('Z (m)')
        plt.axis('scaled')

# Plot the (E, pitch) distribution at a given MC cell
        plt.figure()
        R = self.dressInput['R'][jcell]
        Z = self.dressInput['Z'][jcell]
        F = self.dressInput['F'][jcell].T

        plt.pcolor(self.dressInput['E'], self.dressInput['pitch'], F)
        plt.title(f'D distribution at (R, Z) = ({round(R,2)}, {round(Z,2)}) m')
        plt.xlabel('E (keV)')
        plt.ylabel('pitch')


    def plotSpectra(self):
        '''Plot DRESS output Neutron Emission Spectra'''

        logger.info('Plotting LoS spectrum')
        fig = plt.figure(figsize=(8.8, 5.9))

        ax_nes = fig.add_subplot(1, 2, 1)
        plt.plot(self.En, self.bt, label='BT')
        plt.plot(self.En, self.th, label='TH')
        plt.plot(self.En, self.bb, label='BB')
        ymax = ax_nes.get_ylim()[1]
        plt.plot([2452, 2452], [0, 1.5*ymax], 'k-') # Reference d-d neutron energy
        plt.ylim([0, ymax])
        plt.xlabel('Neutron energy (keV)')
        plt.ylabel('Energy spectrum (neuts/keV/s)')
        if self.los:
            plt.title('Line-of-sight Neutron Emission Spectrum')
        else:
            plt.title('Total Neutron Emission Spectrum')
        plt.legend()
        plt.xlim(2000, 3000)

        ax_phs = fig.add_subplot(1, 2, 2)
        plt.plot(self.phs['Elight_MeVee'], self.phs['bt'], label='BT')
        plt.plot(self.phs['Elight_MeVee'], self.phs['th'], label='TH')
        plt.plot(self.phs['Elight_MeVee'], self.phs['bb'], label='BB')
        plt.xlabel('Light equivalent energy (MeVee)')
        plt.xlim(0, 2)
        ax_phs.set_ylim(bottom=0)

        if self.los:
            plt.title('Line-of-sight Pulse Height Spectrum')
        else:
            plt.title('Total Pulse Height')

        return fig
        

if __name__ == '__main__':

    f_los   = 'input/aug_BC501A.los' # Detector LoS file
# TRANSP: plasma, distribution
    f_pl_tr = 'input/36557D05.CDF'
    f_f_tr  = 'input/36557D05_fi_1.cdf'
#ASCOT output
    f_as    = 'input/29795_3.0s_ascot.h5'

# Total spectra
#    spec = nSpectrum(f_pl_tr, f_f_tr, src='transp', samples_per_volume_element=1e3)
    spec = nSpectrum(f_as, f_as, src='ascot', samples_per_volume_element=1e3)

# LoS spectra
#    spec = nSpectrum(f_pl_tr, f_f_tr, f_los=f_los, src='transp', samples_per_volume_element=1e1)
#    spec = nSpectrum(f_as, f_as, f_los=f_los, src='ascot', samples_per_volume_element=1e3)

    spec.run()

#    spec.fromFile('output/tot_tr_mc1e5.dat')
#    spec.fromFile('output/tot_as_mc1e5.dat')
#    spec.fromFile('output/los_tr_mc1e1.dat')
#    spec.fromFile('output/los_tr_mc1e2.dat')
#    spec.fromFile('output/los_tr_mc1e3.dat')
#    spec.fromFile('output/los_as_mc1e1.dat')
#    spec.fromFile('output/los_as_mc1e2.dat')
#    spec.fromFile('output/los_as_mc1e3.dat')

    spec.plotInput()
    spec.plotSpectra()
    spec.storeSpectra()

    plt.show()



import os, sys, datetime, logging
import numpy as np
from . import rw_for
from scipy.io import netcdf_file

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s | %(message)s', '%Y-%m-%d %H:%M:%S')
logger = logging.getLogger('resp')
logger.setLevel(level=logging.DEBUG)
hnd  = logging.StreamHandler()
hnd.setLevel(level=logging.INFO)
hnd.setFormatter(fmt)
logger.addHandler(hnd)
logger.propagate = False

nereusDir = os.path.dirname(os.path.realpath(__file__))
responseDir = '%s/responses' %nereusDir
out_dir = '%s/output' %nereusDir
os.system('mkdir -p %s' %out_dir)


def gauss_kernel(Emid, sigma, Ec_sim):

    dE = Ec_sim/2.5066

#  Gaussian convolution kernel

    gauss_exp = (Emid[:, None] - Emid[None, :])/sigma[None, :]
    gau_kernel =  dE/sigma[None, :]*np.exp(-0.5*gauss_exp**2)

    return gau_kernel


class RESP:


    def __init__(self):

        pass


    def from_nresp(self, f_spc='%s/SPECT_MPI.DAT' %responseDir):

        self.spc_d = {}
        self.tof_d = {}

        logger.info('Reading file %s' %f_spc)

        f = open(f_spc,'r')
        lines = f.readlines()
        f.close()

        sEbin = lines[0]
        self.Ebin_MeVee = float(sEbin)

        En     = []
        En_wid = []
        self.spc_d  = []
        self.count  = []
        sEn_list = []

        jEn = -1
        for lin in lines[1:]:
            slin = lin.strip()
            if (slin == ''):
                continue
            sarr = slin.split()
            try:
                tmp = float(sarr[0])
                for snum in sarr:
                    self.spc_d[jEn][lbl].append(float(snum))
            except:
                lbl, sEn, sEn_wid, snr = slin.split()
                if sEn not in sEn_list:
                    jEn += 1
                    En.append(float(sEn))
                    En_wid.append(float(sEn_wid))
                    self.spc_d.append({})
                    self.count.append({})
                    sEn_list.append(sEn)
                self.spc_d[jEn][lbl] = []
                if lbl[:5] == 'PP3AS':
                    self.count[jEn][lbl] = 0
                else:
                    self.count[jEn][lbl] = int(snr)

        self.En_MeV     = 1e-3*np.array(En    , dtype=np.float32)
        self.En_wid_MeV = 1e-3*np.array(En_wid, dtype=np.float32)
        nEn = len(self.En_MeV)
        self.phs_max = 0
        for j in range(nEn):
            for lbl, spec in self.spc_d[j].items():
                self.phs_max = max(self.phs_max, len(spec))

        self.EphsB_MeVee = self.Ebin_MeVee*np.arange(self.phs_max + 1)
        self.Ephs_MeVee = 0.5*(self.EphsB_MeVee[1:] + self.EphsB_MeVee[:-1])
        self.RespMat = np.zeros((nEn, self.phs_max))

        for jEn in range(nEn):
            for lbl, arr in self.spc_d[jEn].items():
                if lbl[:5] != 'PP3AS':
                    myarr = np.zeros(self.phs_max)
                    nloc = len(arr)
                    myarr[:nloc] = arr[:nloc]
                    self.RespMat[jEn, :] += myarr


    def from_cdf(self, f_cdf):

        logger.info('Reading file %s' %f_cdf)

        cv = netcdf_file(f_cdf, 'r', mmap=False).variables

        self.Ebin_MeVee  = cv['Ebin'][:]
        self.RespMat     = cv['ResponseMatrix'].data
        self.En_MeV      = cv['E_NEUT'][:]
        self.En_wid_MeV  = cv['En_wid'][:]
        self.EphsB_MeVee = cv['E_light_B'][:]
        self.Ephs_MeVee  = cv['E_light'][:]


    def from_hepro(self, f_hep='%s/simresp.rsp' %responseDir):

        logger.info('Reading file %s' %f_hep)

        f = open(f_hep, 'r')

        data = []
        for line in f.readlines():
            data += line.split()
        f.close()

        dsim = np.array(data, dtype = np.float32)
        n_dsim = len(dsim)
        logger.debug('N_DSIM %d', n_dsim)
        ndim = []
        dE = []
        en1 = []
        en2 = []
        spc = []
        self.Ebin_MeVee = dsim[0]
        j = 1
        while j < n_dsim:
            dE.append(dsim[j])
            nsize = int(dsim[j+1])
            ndim.append(nsize)
            en1.append(dsim[j+2])
            en2.append(dsim[j+3])
            j += 4
            logger.debug('nsize %d %d', nsize, j)
            spc.append(dsim[j:j+nsize])
            j += nsize
        self.En_MeV = np.array(dE, dtype=np.float32)
        en1  = np.array(en1, dtype=np.float32)
        en2  = np.array(en2, dtype=np.float32)
        self.En_wid_Mev = en2 - en1
        ndims = np.array(ndim, dtype=np.int32)
        nEn = len(self.En_MeV)
        self.phs_max = np.max(ndims)
        self.RespMat = np.zeros((nEn, self.phs_max), dtype=np.float32)
        for jEn, phs in enumerate(spc):
            nphs = len(phs)
            self.RespMat[jEn, :nphs] = phs[:nphs]
        self.EphsB_MeVee = self.Ebin_MeVee*np.arange(self.phs_max + 1)
        self.Ephs_MeVee = 0.5*(self.EphsB_MeVee[1:] + self.EphsB_MeVee[:-1])


    def broaden(self, f_par='%s/neut_fit.txt' %responseDir):
        '''Gaussian broadening of the response function'''

# Broadening parameters

        logger.info('Gaussian convolution')
        f = open(f_par, 'r')
        lin = f.readlines()[13].split()
        a = float(lin[3])
        b = float(lin[4])
        c = float(lin[5])
        z = float(lin[6])
        Ec_fac = float(lin[7])
        logger.debug('%8.4f %8.4f %8.4f %8.4 %8.4f', a, b, c, z, Ec_fac)

        Emid = self.Ephs_MeVee
        sigma = np.sqrt(a**2 * Emid**2 + b**2 * Emid + c**2)/235.48

        logger.info('Gaussian kernel from file %s', f_par)
        gau_ker = gauss_kernel(self.Ephs_MeVee, sigma, self.Ebin_MeVee)
        self.RespMat_gb = np.einsum('ij,jk->ik', self.RespMat, gau_ker)


    def to_hepro(self, fout='%s/ddnpar.asc' %responseDir):

        f = open(fout, 'w')
        f.write('  %13.7e\n' %self.Ebin_MeVee)
        n_En, n_spc = self.RespMat.shape
        for jEn in range(n_En):
            f.write('  %11.5f       %5d  %11.5f  %11.5f\n' %(self.En_MeV[jEn], n_spc, self.EphsB_MeVee[0], self.EphsB_MeVee[-1]))
            spc_str = rw_for.wr_for(self.RespMat[jEn, :], fmt=' %13.6e', n_lin=6)
            f.write(spc_str)
        f.close()
        logger.info('Written %s' %fout)


    def to_cdf(self, f_cdf=None):

        if f_cdf is None:
            f_cdf = '%s/rm.cdf' %out_dir
        fcdf = f_cdf
        jcdf = 0
        while os.path.isfile(fcdf):
            jcdf += 1
            fcdf = '%s%s' %(f_cdf, jcdf)

        len_pos = []
        nEn = len(self.En_MeV)
        for jEn in range(nEn):
            phs = self.RespMat[jEn, :]
            if phs.any() > 0:
                (ind_pos, ) = np.where(phs > 0.)
                len_pos.append(ind_pos[-1])
            else:
                len_pos.append(0)
        jEn_max = np.argmax(len_pos)
        nEp = min( len_pos[jEn_max] + 1, len(self.RespMat[jEn_max, :]) ) # can get down by one

# NetCDF output

        f = netcdf_file(fcdf, 'w', mmap=False)

        f.history = "Created " + datetime.datetime.today().strftime("%d/%m/%y")

        f.createDimension('E_NEUT'   , nEn)
        f.createDimension('E_light'  , nEp)
        f.createDimension('E_light_B', nEp + 1)
        f.createDimension('Ebin_dim' , 1)

        En = f.createVariable('E_NEUT', np.float32, ('E_NEUT', ))
        En[:] = self.En_MeV
        En.units = 'MeV'
        En.long_name = 'Neutron energy'

        Ewid = f.createVariable('En_wid', np.float32, ('E_NEUT', ))
        Ewid[:] = self.En_wid_MeV
        Ewid.units = 'MeV'
        Ewid.long_name = 'NRESP-energy width for each En'

        Ep = f.createVariable('E_light', np.float32, ('E_light', ))
        Ep[:] = self.Ephs_MeVee[:nEp]
        Ep.units = 'MeVee'
        Ep.long_name = 'Equivalent photon energy grid'

        EpB = f.createVariable('E_light_B', np.float32, ('E_light_B', ))
        EpB[:] = self.EphsB_MeVee[:nEp+1]
        EpB.units = 'MeVee'
        EpB.long_name = 'Equivalent photon energy bins'

        Ebin = f.createVariable('Ebin', np.float32, ('Ebin_dim', ))
        Ebin[:] = self.Ebin_MeVee
        Ebin.units = 'MeVee'
        Ebin.long_name = 'Step for PHS bins'

        rm = f.createVariable('ResponseMatrix', np.float32, ('E_NEUT', 'E_light'))
        rm.units = '1/(s MeVee)'
        rm.long_name = 'Response functions for several neutron energies'
        rm[:] = self.RespMat[:, :nEp]

        f.close()
        logger.info('Stored %s' %fcdf)


if __name__ == "__main__":

    import matplotlib.pylab as plt

    f_nresp = '%s/SPECT_MPI.DAT' %responseDir
#    f_nresp = 'nresp.dat'
    f_cdf = 'rm2.cdf'

# Store NetCDF
    rsp = RESP()
    rsp.from_nresp(f_spc=f_nresp)
    rsp.to_cdf(f_cdf)
# Store HEPRO output for TOFANA
    rsp.to_hepro()

# Gaussian broadening, storing as cdf
    rsp.from_cdf(f_cdf)
    rsp.broaden()
    rsp.RespMat = rsp.RespMat_gb

# Store gaussian-broadened into NetCDF
    fcdf_out = 'rm_bg.cdf'
    rsp.to_cdf(fcdf_out)

# Plot a few functions

    cv1 = netcdf_file('rm2.cdf'  , 'r', mmap=False).variables
    cv2 = netcdf_file('rm_bg.cdf', 'r', mmap=False).variables

    Emid = cv1['E_light'][:]
    En = cv1['E_NEUT'][:]
    jEn = 0
    plt.plot(Emid, cv1['ResponseMatrix'][jEn] , 'r-', label='En = %5.3f MeV'     %En[jEn])
    plt.plot(Emid, cv2['ResponseMatrix'][jEn] , 'g-', label='En = %5.3f MeV, GB' %En[jEn])
    jEn = -10
    plt.plot(Emid, cv1['ResponseMatrix'][jEn], 'm-', label='En = %5.3f MeV'     %En[jEn])
    plt.plot(Emid, cv2['ResponseMatrix'][jEn], 'b-', label='En = %5.3f MeV, GB' %En[jEn])

    plt.xlim([0, 5])
    plt.ylim([0, 1])
    plt.legend()

    plt.show()



import sys
import matplotlib
import numpy as np

try:
    from PyQt5.QtWidgets import QWidget, QTabWidget, QVBoxLayout
    from PyQt5.QtCore import QRect
    matplotlib.use('Qt5Agg')
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
except:
    from PyQt4.QtGui import QWidget, QTabWidget, QVBoxLayout
    from PyQt4.QtCore import QRect
    matplotlib.use('Qt4Agg')
    from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar

import matplotlib.pylab as plt

fsize   = 8
titsize = 10
lblsize = 10
fig_size = (8.8, 5.9)
lwid = 2.5


class plotWindow(QWidget):


    def __init__(self, geom=(80, 30, 900, 680)):

        if sys.version_info[0] == 3:
            super().__init__()
        else:
            super(QWidget, self).__init__()
        self.setGeometry(QRect(*geom))
        self.tabs = QTabWidget(self)
        self.tabs.setStyleSheet("QTabBar::tab { width: 165 }")
        self.setWindowTitle('Nuclear reactions properties')


    def addPlot(self, title, figure):

        new_tab = QWidget()
        layout = QVBoxLayout()
        new_tab.setLayout(layout)

        new_canvas = FigureCanvas(figure)
        new_toolbar = NavigationToolbar(new_canvas, new_tab)
        layout.addWidget(new_canvas)
        layout.addWidget(new_toolbar)
        for jtab in range(self.tabs.count()):
            if self.tabs.tabText(jtab) == title:
                self.tabs.removeTab(jtab)
        self.tabs.addTab(new_tab, title)
        self.tabs.setCurrentIndex(self.tabs.count()-1)


def fig_reactivity(reac_d, Ti_keV, color='#d0d0d0', figSize=None):

    if figSize is None:
        figSize = fig_size
    fig_reac = plt.figure('Reac', figSize, dpi=100, facecolor=color)
    plt.cla()
    for lbl, reac in reac_d.items():
#        if lbl[:3] != 'D(D':
            plt.loglog(Ti_keV, reac, label=lbl, linewidth=lwid)
    dd = reac_d['D(D,n)3He'] + reac_d['D(D,P)T']
#    plt.loglog(Ti_keV, dd, label='DD', linewidth=lwid)
    plt.xlim([Ti_keV[0], Ti_keV[-1]])
    plt.ylim([1e-19, 1e-14])
    plt.xlabel(r'T$_i$ [keV]')
    plt.ylabel(r'Reactivity [cm$^3$/s]')
    plt.grid()
    plt.legend(loc=4)

    return fig_reac


def fig_cross(reac, theta, Egrid, log_scale=False, color='#d0d0d0'):

    fig_cs = plt.figure('Cross-sections', fig_size, dpi=100, facecolor=color)
    plt.cla()
    the_deg = np.degrees(theta)
    for j, E in enumerate(Egrid):
        if log_scale:
            plt.semilogy(the_deg, reac[:, j], linewidth=lwid, label='E=%5.3f MeV' %E)
        else:
            plt.plot(the_deg, reac[:, j], linewidth=lwid, label='E=%5.3f MeV' %E)
    plt.xlim([the_deg[0], the_deg[-1]])
    plt.legend()
    plt.xlabel('Scattering angle [deg]')
    plt.ylabel(r'$\frac{d\sigma}{d\Omega}$ [mbarn]')

    return fig_cs


def fig_resp(resp, En_MeV=2.5, color='#d0d0d0'):

    d2 = (resp.En_MeV - En_MeV)**2
    jEn = np.argmin(d2)
    fig_resp = plt.figure('Response', (8.8, 5.9), dpi=100)
    plt.cla()
    plt.plot(resp.Ephs_MeVee, resp.RespMat[jEn], 'r-', label='En=%5.3f MeV' %resp.En_MeV[jEn], linewidth=lwid)
    if hasattr(resp, 'RespMat_gb'):
        plt.plot(resp.Ephs_MeVee, resp.RespMat_gb[jEn], 'b-', label='En=%5.3f MeV, GB' %resp.En_MeV[jEn], linewidth=lwid)
    plt.legend()
    plt.xlim([0, 0.5*En_MeV])
    plt.ylim([0, 0.3/En_MeV])
    plt.xlabel('Pulse Height [MeVee]')
    plt.ylabel('Pulse Height Spectrum')

    return fig_resp


def fig_spec(Egrid, Espec, color='#d0d0d0'):

    fig_spc = plt.figure('Spectrum', fig_size, dpi=100, facecolor=color)
    plt.cla()
#    plt.semilogy(Egrid, Espec)
    plt.plot(Egrid, Espec, linewidth=lwid)
    plt.xlabel('En [MeV]')
    plt.ylabel('dN/dEn')

    return fig_spc



#!/usr/bin/env python

import os, logging
import numpy as np
import matplotlib.pylab as plt
from multiprocessing import Pool, cpu_count
from nresp.en2light import En2light
from nresp import crossSections
import rw_for

nrespDir = os.path.dirname(os.path.realpath(__file__))
os.system('mkdir -p %s/output' %nrespDir)

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s | %(message)s', '%Y-%m-%d %H:%M:%S')
logger = logging.getLogger('nresp')
logger.setLevel(level=logging.DEBUG)
hnd  = logging.StreamHandler()
flog = '%s/output/nresp.log' %nrespDir
fhnd = logging.FileHandler(flog, mode='w') # mode='a' for appending
hnd.setLevel(level=logging.INFO)
fhnd.setLevel(level=logging.DEBUG)
hnd.setFormatter(fmt)
fhnd.setFormatter(fmt)
logger.addHandler(hnd)
logger.addHandler(fhnd)
logger.propagate = False

CS = crossSections.crossSections()
#CS.plot()
flt_typ = np.float64
int_typ = np.int32


class NRESP:


    def __init__(self, nresp_set, parallel=True):

        self.reac_names = [x for x in CS.reacTotUse]
        self.reac_names.append('light-guide')
        self.En_MeV = np.atleast_1d(eval(nresp_set['Energy array']))
        self.nresp_set = nresp_set
        self.nEn = len(self.En_MeV)
        jmax = np.argmax(self.En_MeV)
        self.n_react = len(self.reac_names)
        self.En_wid_MeV = self.nresp_set['En_wid_frac']*self.En_MeV
        En_wid = self.En_wid_MeV[jmax]
        if self.nresp_set['distr'] == 'mono':
            self.phs_max = int(self.En_MeV[jmax]/self.nresp_set['Ebin_MeVee'])
        elif self.nresp_set['distr'] == 'gauss':
            self.phs_max = int((self.En_MeV[jmax] + 5*En_wid)/self.nresp_set['Ebin_MeVee'])
        self.EphsB_MeVee = self.nresp_set['Ebin_MeVee']*np.arange(self.phs_max + 1)
        self.Ephs_MeVee = 0.5*(self.EphsB_MeVee[1:] + self.EphsB_MeVee[:-1])

        if parallel:
            self.run_multi()
        else:
            self.run_serial()


    def run_multi(self):

        timeout_pool = int(self.nresp_set['nmc']/1e3)
        pool = Pool(cpu_count())
        out = pool.map_async(En2light, [(EMeV, self.phs_max, self.nresp_set) for EMeV in self.En_MeV]).get(timeout_pool)
        pool.close()
        pool.join()

        logger.info('END light output calculation, nMC=%d, nEn=%d', self.nresp_set['nmc'], len(self.En_MeV))
        self.count_reac   = np.zeros((self.nEn, self.n_react), dtype=int_typ)
        self.phs_dim_rea  = np.zeros((self.nEn, self.n_react), dtype=int_typ)
        self.count_pp3as  = np.zeros((self.nEn, CS.max_level), dtype=int_typ)
        self.phs_dim_pp3  = np.zeros((self.nEn, CS.max_level), dtype=int_typ)
        self.pp3as_output = np.zeros((self.nEn, CS.max_level, self.phs_max), dtype=flt_typ)
        self.light_output = np.zeros((self.nEn, self.n_react, self.phs_max), dtype=flt_typ)
        for jE, x in enumerate(out):
            self.count_reac  [jE] = x[0]
            self.count_pp3as [jE] = x[1]
            self.phs_dim_rea [jE] = x[2]
            self.phs_dim_pp3 [jE] = x[3]
            self.light_output[jE] = x[4]
            self.pp3as_output[jE] = x[5]
        self.phs_dim_rea += 1
        self.phs_dim_pp3 += 1
        self.RespMat = np.sum(self.light_output, axis=1)


    def run_serial(self):

        self.count_reac   = np.zeros((self.nEn, self.n_react), dtype=int_typ)
        self.phs_dim_rea  = np.zeros((self.nEn, self.n_react), dtype=int_typ)
        self.count_pp3as  = np.zeros((self.nEn, CS.max_level), dtype=int_typ)
        self.phs_dim_pp3  = np.zeros((self.nEn, CS.max_level), dtype=int_typ)
        self.pp3as_output = np.zeros((self.nEn, CS.max_level, self.phs_max), dtype=flt_typ)
        self.light_output = np.zeros((self.nEn, self.n_react, self.phs_max), dtype=flt_typ)
        for jE, EMeV in enumerate(self.En_MeV):
            self.count_reac[jE], self.count_pp3as[jE], self.phs_dim_rea[jE], self.phs_dim_pp3[jE], \
                self.light_output[jE], self.pp3as_output[jE] = En2light((EMeV, self.phs_max, self.nresp_set))
        self.phs_dim_rea += 1
        self.phs_dim_pp3 += 1
        self.RespMat = np.sum(self.light_output, axis=1)


    def to_nresp(self, fout='%s/output/spect.dat' %nrespDir):

        f = open(fout, 'w')
        f.write('%15.6e\n' %self.nresp_set['Ebin_MeVee'])
        for jEn, En in enumerate(self.En_MeV):
            for jreac, reac in enumerate(self.reac_names):
                count = self.count_reac[jEn, jreac]
                if count > 0:
                    phs_dim = self.phs_dim_rea[jEn, jreac]
                    f.write('%-30s %8.2f %8.2f %13d\n' %(reac, 1e3*En, 1e3*self.En_wid_MeV[jEn], count))
                    spc_str = rw_for.wr_for(self.light_output[jEn, jreac, :phs_dim], fmt=' %13.6e', n_lin=5)
                    f.write(spc_str)

            for jlevel in range(CS.max_level):
                count = self.count_pp3as[jEn, jlevel]
                if count > 0:
                    phs_dim = self.phs_dim_pp3[jEn, jlevel]
                    lbl = 'PP3AS%d' %(jlevel+1)
                    f.write('%-30s %8.2f %8.2f %13d\n' %(lbl, En, self.En_wid_MeV[jEn], count))
                    spc_str = rw_for.wr_for(self.pp3as_output[jEn, jlevel, :phs_dim], fmt=' %13.6e', n_lin=5)
                    f.write(spc_str)
        f.close()
        logger.info('Written %s' %fout)


    def plotResponse(self, E_MeV=2.):

        jEn = np.argmin(np.abs(self.En_MeV - E_MeV))
        n_react = self.light_output.shape[1]

        fig = plt.figure('NRESP reactions', (8.8, 5.9))
        fig.clf()

        fig.text(0.5, 0.95, r'E$_n$=%6.3f MeV' %self.En_MeV[jEn], ha='center')
        ax1 = fig.add_subplot(1, 2, 1)
        for jreact in range(n_react):
            ax1.plot(self.Ephs_MeVee, self.light_output[jEn, jreact], label=self.reac_names[jreact])
        ax1.set_ylim([0., 20./E_MeV])
        ax1.set_xlabel('Pulse Height [MeVee]')
        ax1.set_ylabel('Pulse Height Spectrum')
        ax1.legend()

        ax2 = fig.add_subplot(1, 2, 2)
        resp = np.sum(self.light_output, axis=1)
        ax2.plot(self.Ephs_MeVee, resp[jEn])
        ax2.set_ylim([0., 100./E_MeV])
        ax2.set_xlabel('Pulse Height [MeVee]')
        ax2.set_ylabel('Pulse Height Spectrum')
        return fig



#!/usr/bin/env python

import os, sys, logging, webbrowser, json

try:
    from PyQt5.QtWidgets import QMainWindow, QWidget, QApplication, QGridLayout, QMenu, QAction, QLabel, QPushButton, QLineEdit, QCheckBox, QFileDialog, QRadioButton, QButtonGroup, QTabWidget, QVBoxLayout, QComboBox, QSpinBox, QDoubleSpinBox
    from PyQt5.QtGui import QPixmap, QIcon
    from PyQt5.QtCore import Qt, QRect, QSize, QLocale
    qt5 = True
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
except:
    from PyQt4.QtCore import Qt, QRect, QSize, QLocale
    from PyQt4.QtGui import QPixmap, QIcon, QMainWindow, QWidget, QApplication, QGridLayout, QMenu, QAction, QLabel, QPushButton, QLineEdit, QCheckBox, QFileDialog, QRadioButton, QButtonGroup, QTabWidget, QVBoxLayout, QComboBox, QSpinBox, QDoubleSpinBox
    from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
    qt5 = False

import numpy as np
import matplotlib.pylab as plt
from mpl_toolkits.mplot3d import Axes3D
from .reactivities import *
import .calc_cross_section as cs
from .los import los
from . import plots, response
from .reactions import reaction
from .nresp import nresp
from .dress_client import nspectrum

os.environ['BROWSER'] = '/usr/bin/firefox'

usLocale = QLocale('us')

logger = logging.getLogger('nereus.gui')
logger.setLevel(logging.DEBUG)

nereusDir = os.path.dirname(os.path.realpath(__file__))

os.system('mkdir -p %s/nresp/output' %nereusDir)
os.system('mkdir -p %s/dress_client/output' %nereusDir)


class NEREUS(QMainWindow):


    def __init__(self):

        if sys.version_info[0] == 3:
            super().__init__()
        else:
            super(QMainWindow, self).__init__()

        self.setLocale(usLocale)

        xwin  = 1000
        yhead = 44
        yline = 30
        ybar  = 54
        yicon = 46
        xicon = yicon
        ywin  = 15*yline + yhead + ybar

        qhead  = QWidget(self)
        qbar   = QWidget(self)
        qtabs  = QTabWidget(self)
        qhead.setGeometry(QRect(0,     0, xwin, yhead))
        qbar.setGeometry(QRect(0, yhead, xwin, ybar))
        qtabs.setGeometry(QRect(0, yhead+ybar+10, xwin, ywin-yhead-ybar-10))
        qtabs.setStyleSheet("QTabBar::tab { width: 160 }")
        header_grid = QGridLayout(qhead) 
        tbar_grid   = QGridLayout(qbar) 

#-----
# Tabs
#-----

        qreac = QWidget()
        reac_layout = QGridLayout()
        qreac.setLayout(reac_layout)
        qtabs.addTab(qreac, 'Reactivities')

        qcross = QWidget()
        cross_layout = QGridLayout()
        qcross.setLayout(cross_layout)
        qtabs.addTab(qcross, 'Cross-sections')

        qresp = QWidget()
        resp_layout = QGridLayout()
        qresp.setLayout(resp_layout)
        qtabs.addTab(qresp, 'Response Matrix')

        qspec = QWidget()
        spec_layout = QGridLayout()
        qspec.setLayout(spec_layout)
        qtabs.addTab(qspec, 'Neutron spectra')

        qlos = QWidget()
        los_layout = QGridLayout()
        qlos.setLayout(los_layout)
        qtabs.addTab(qlos, 'Detector LoS')

        qnresp = QWidget()
        nresp_layout = QGridLayout()
        qnresp.setLayout(nresp_layout)
        qtabs.addTab(qnresp, 'NRESP')

#--------
# Menubar
#--------

        menubar = self.menuBar()
        fileMenu = QMenu('&File', self)
        jsonMenu  = QMenu('&Setup', self)
        helpMenu = QMenu('&Help', self)
        menubar.addMenu(fileMenu)
        menubar.addMenu(jsonMenu)
        menubar.addMenu(helpMenu)

        fmap = {'reac': self.reactivity, 'csec': self.cross_section, 'resp': self.response, \
                'spec': self.spectra, 'los': self.los, 'nresp': self.nresp, 'exit': sys.exit}
        qlbl = {'reac': '&Reactivity', 'csec': '&Cross-section', 'resp': '&Response Matrix', \
                'spec': '&Spectra', 'los': '&Line-of-Sight', 'nresp': '&NRESP', 'exit': 'Exit'}
        Action = {}
        for key, lbl in qlbl.items():
            Action = QAction(lbl, fileMenu)
            Action.triggered.connect(fmap[key])
            if key == 'exit':
                fileMenu.addSeparator()
            fileMenu.addAction(Action)

        loadAction  = QAction('&Load Setup', jsonMenu)
        saveAction  = QAction('&Save Setup', jsonMenu)
        jsonMenu.addAction(loadAction)
        jsonMenu.addAction(saveAction)
        loadAction.triggered.connect(self.loadJSON)
        saveAction.triggered.connect(self.saveJSON)

        aboutAction = QAction('&Web docu', helpMenu)
        aboutAction.triggered.connect(self.about)
        helpMenu.addAction(aboutAction)

        header_grid.addWidget(menubar, 0, 0, 1, 10)

# Icons
        dum_lbl  = QLabel(200*' ')
        xpos = 0
        ypos = 0
        for jpos, key in enumerate(qlbl.keys()):
            but = QPushButton()
            but.setGeometry(xpos, xicon, ypos, yicon)
            but.setStyleSheet("min-height: %dpx; padding: 0.0em 0.0em 0.0em 0.0em" %yicon)
            fgif = '%s/%s.gif' %(nereusDir, key)
            if os.path.isfile(fgif):
                but.setIcon(QIcon(fgif))
            else:
                fpng = '%s/%s.png' %(nereusDir, key)
                but.setIcon(QIcon(fpng))
            but.setIconSize(QSize(ybar, ybar))
            but.clicked.connect(fmap[key])
            tbar_grid.addWidget(but, 0, jpos)
            xpos += yicon + 4

        tbar_grid.addWidget(dum_lbl, 0, jpos+1, 1, 10)

# User options

        f_json = '%s/settings/default.json' %nereusDir
        with open(f_json, 'r') as fjson:
            self.setup_init = json.load(fjson)
        self.gui = {}
        for node in self.setup_init.keys():
            self.gui[node] = {}
        user = os.getenv('USER')

#------------
# Reactivity
#------------

        cb = ['D(D,n)3He', 'D(D,P)T', 'D(T,n)4He', 'D(3He,P)4He']
        self.fill_layout(reac_layout, 'reac', checkbuts=cb)

#---------------
# Cross-sections
#---------------

        entries = ['E', 'Z1', 'Z2']
        cb = ['log_scale']
        combos = {'reac': reaction.keys()}
        self.fill_layout(cross_layout, 'cross', entries=entries, checkbuts=cb, combos=combos)

#---------
# Response
#---------

        entries = ['Input file', 'Gaussian broadening', 'Plot Eneut [MeV]']
        combos = {'Write response': ['None', 'CDF', 'hepro']}
        self.fill_layout(resp_layout, 'response', entries=entries, combos=combos, lbl_wid=140, ent_wid=360)

#--------
# Spectra
#--------

        entries = ['TRANSP plasma', 'TRANSP fast ions', 'ASCOT file', 'Detector LoS', '#MonteCarlo', 'Response file', 'Output file']
        combos = {'Code': ['TRANSP', 'ASCOT'], 'Spectrum': ['Total', 'Line-of-sight']}
        cb = ['Store spectra', 'MultiProcess'] 
        self.fill_layout(spec_layout, 'spectrum', entries=entries, combos=combos, checkbuts=cb, ent_wid=360)

#---------
# LOS
#---------

        entries = ['disk_thick', 'cell_radius', 'coll_diam', 'd_det_coll', 'det_radius', 'tilt',
            'tan_radius', 'y_det', 'z_det', 'Rmaj', 'r_chamb', 'label']
        cb = ['Write LOS']
        self.fill_layout(los_layout, 'detector', entries=entries, checkbuts=cb)

#---------
# NRESP
#---------

        entries = ['Energy array', 'f_detector', 'f_in_light', 'nmc', 'En_wid_frac', 'Ebin_MeVee', 'Energy for PHS plot']
        combos = {'distr': ['gauss', 'mono']}
        cb = ['Write nresp', 'MultiProcess']
        self.fill_layout(nresp_layout, 'nresp', entries=entries, combos=combos, checkbuts=cb)

#-----------
# GUI layout
#-----------

        self.setWindowTitle('NEREUS')
        icon = QIcon()
        icon.addPixmap(QPixmap('%s/nereus_icon.png' %nereusDir), QIcon.Selected, QIcon.On)
        self.setWindowIcon(icon)
        self.setStyleSheet("QLabel { width: 4 }")
        self.setStyleSheet("QLineEdit { width: 4 }")
        self.setGeometry(10, 10, xwin, ywin)
        self.show()


    def about(self):

        webbrowser.open('https://github.com/tardini/neut_reac/wiki')


    def fill_layout(self, layout, node, entries=[], checkbuts=[], combos={}, lbl_wid=140, ent_wid=180, col_shift=0):
# Checkbutton

        jrow = 0

        for key in checkbuts:
            self.gui[node][key] = QCheckBox(key)
            layout.addWidget(self.gui[node][key], jrow, col_shift)
            if self.setup_init[node][key]:
                self.gui[node][key].setChecked(True)
            jrow += 1

        for key in entries:
            val = self.setup_init[node][key]
            qlbl = QLabel(key)
            qlbl.setFixedWidth(lbl_wid)
            if isinstance(val, int):
                self.gui[node][key] = QSpinBox()
                self.gui[node][key].setRange(0, int(1e9))
                self.gui[node][key].setValue(val)
            elif isinstance(val, float):
                self.gui[node][key] = QDoubleSpinBox()
                self.gui[node][key].setRange(-1000., 1000.)
                self.gui[node][key].setDecimals(4) 
                self.gui[node][key].setValue(val)
            else:
                self.gui[node][key] = QLineEdit(str(val))
            self.gui[node][key].setFixedWidth(ent_wid)
            layout.addWidget(qlbl               , jrow, col_shift)
            layout.addWidget(self.gui[node][key], jrow, col_shift+1)
            jrow += 1

        for key, combs in combos.items():
            qlbl = QLabel(key)
            qlbl.setFixedWidth(lbl_wid)
            self.gui[node][key] = QComboBox()
            for comb in combs:
                self.gui[node][key].addItem(comb.strip())
            index = self.gui[node][key].findText(self.setup_init[node][key].strip())
            self.gui[node][key].setCurrentIndex(index)
            layout.addWidget(qlbl               , jrow, col_shift)
            layout.addWidget(self.gui[node][key], jrow, col_shift+1)
            jrow += 1

        layout.setRowStretch(layout.rowCount(), 1)
        layout.setColumnStretch(layout.columnCount(), 1)


    def gui2json(self):
        '''Returns a dic'''
        json_d = {}
        for node in self.gui.keys():
            json_d[node] = self.get_gui_tab(node)
        return json_d


    def get_gui_tab(self, node):

        node_dic = {}
        for key, val in self.gui[node].items():
            node_dic[key] = {}
            if isinstance(val, QLineEdit):
                node_dic[key] = val.text()
            elif isinstance(val, QSpinBox):
                node_dic[key] = int(val.text())
            elif isinstance(val, QDoubleSpinBox):
                node_dic[key] = float(val.text())
            elif isinstance(val, QCheckBox):
                node_dic[key] = val.isChecked()
            elif isinstance(val, QComboBox):
                node_dic[key] = val.currentText()
            elif isinstance(val, QButtonGroup):
                bid = val.checkedId()
                node_dic[key] = self.rblists[key][bid]

        return node_dic


    def set_gui(self, json_d):

        for node, val1 in json_d.items():
            for key, vald in val1.items():
                if key not in self.gui[node].keys():
                    continue
                widget = self.gui[node][key]
                if isinstance(widget, QCheckBox):
                    widget.setChecked(vald)
                elif isinstance(widget, QButtonGroup):
                    for but in widget.buttons():
                        if but.text() == vald:
                            but.setChecked(True)
                elif isinstance(widget, QLineEdit):
                    if vald:
                        widget.setText(str(vald))
                elif isinstance(widget, QComboBox):
                    for index in range(widget.count()):
                        if widget.itemText(index).strip() == vald.strip():
                            widget.setCurrentIndex(index)
                            break


    def loadJSON(self):

        ftmp = QFileDialog.getOpenFileName(self, 'Open file', \
            '%s/settings' %nereusDir, "json files (*.json)")
        if qt5:
            f_json = ftmp[0]
        else:
            f_json = str(ftmp)

        with open(f_json) as fjson:
            setup_d = json.load(fjson)
        self.set_gui(setup_d)


    def saveJSON(self):

        out_dic = self.gui2json()
        ftmp = QFileDialog.getSaveFileName(self, 'Save file', \
            '%s/settings' %nereusDir, "json files (*.json)")
        if qt5:
            f_json = ftmp[0]
        else:
            f_json = str(ftmp)
        with open(f_json, 'w') as fjson:
            fjson.write(json.dumps(out_dic))


    def reactivity(self):

        reac_dic = self.get_gui_tab('reac')

        Ti_keV = np.linspace(1., 1000., 1000)
        
        logger.info('Reactivities')

        reac_d = {}

        for key, reac in reaction.items():
            if hasattr(reac, 'coeff_reac') and reac_dic[key]:
                reac_d[key] = react(Ti_keV, key)

        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        fig_reac =  plots.fig_reactivity(reac_d, Ti_keV)
        self.wid.addPlot('Reactivities', fig_reac)
        self.wid.show()


    def cross_section(self):

        cross_dic = self.get_gui_tab('cross')

        Egrid = np.array([float(x) for x in eval(cross_dic['E'])], dtype=np.float32)
        logger.info('Cross-section')

        theta = np.linspace(0, np.pi, 61)
        mu_grid = np.cos(theta)
        reac_lbl = cross_dic['reac']
        sigma = cs.sigma_diff(Egrid, mu_grid, reac_lbl, 2, 2)

        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        fig_cross = plots.fig_cross(sigma, theta, Egrid, log_scale=cross_dic['log_scale'])
        self.wid.addPlot('Cross-sections', fig_cross)
        self.wid.show()


    def response(self):

        resp_d = self.get_gui_tab('response')
        logger.info('Response Matrix')
        out_lbl = resp_d['Write response'].lower()
        f_in = resp_d['Input file']
        f_gb = resp_d['Gaussian broadening']
        En   = resp_d['Plot Eneut [MeV]'] 

        resp = response.RESP()
        fname = ''.join(os.path.splitext(f_in)[:-1])
        ext = os.path.splitext(f_in)[-1]
        print(fname + ext)
        if ext == '.cdf':
            resp.from_cdf(f_in)
        elif ext == '.hepro':
            resp.from_hepro(f_in)
        elif ext in ('.DAT', '.rsp'):
            resp.from_nresp()

        if f_gb:
            resp.broaden()
            fgb_out = '%s_gb.%s' %(fname, out_lbl)

        f_out = '%s.%s' %(fname, out_lbl)
        if out_lbl == 'cdf':
            resp.to_cdf(f_out)
            if f_gb:
                tmp = resp.RespMat
                resp.RespMat = resp.RespMat_gb
                resp.to_cdf(fgb_out)
                resp.RespMat = tmp
        elif out_lbl == 'hepro':
            resp.to_hepro(f_out)
            if f_gb:
                tmp = resp.RespMat
                resp.RespMat = resp.RespMat_gb
                resp.to_hepro(fgb_out)
                resp.RespMat = tmp

        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        print('En=',En)
        fig_resp = plots.fig_resp(resp, En_MeV=En)
        self.wid.addPlot('Response matrix', fig_resp)
        self.wid.show()


    def spectra(self):

        nes_d = self.get_gui_tab('spectrum')

        logger.info('Spectra')
        n_samples = nes_d['#MonteCarlo']
        code = nes_d['Code']

        if code == 'TRANSP':
            f1 = nes_d['TRANSP plasma']
            f2 = nes_d['TRANSP fast ions']
        elif code == 'ASCOT':
            f1 = nes_d['ASCOT file']
            f2 = nes_d['ASCOT file']
        if nes_d['Spectrum'] == 'Total':
            nes = nspectrum.nSpectrum(f1, f2, src=code.lower(), samples_per_volume_element=n_samples)
        elif nes_d['Spectrum'] == 'Line-of-sight':
            nes = nspectrum.nSpectrum(f1, f2, f_los=nes_d['Detector LoS'], src=code.lower(), samples_per_volume_element=n_samples)

        reComp = False
        if not hasattr(self, 'nes_d'):
            reComp = True
        else:
            for key, val in self.nes_d.items():
                if key not in ('Store spectra'):
                    if nes_d[key] != val:
                        reComp = True
                        break
        if reComp:
            nes.run(parallel=nes_d['MultiProcess'])
        else:
            nes = self.nes
        logger.info('Convolving into Pulse Height Spectrum')
        nes.nes2phs(f_resp=nes_d['Response file'])

        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        fig_nes = nes.plotSpectra()
        self.wid.addPlot('Neutron Spectrum', fig_nes)
        self.wid.show()
        self.nes_d = {key: val for key, val in nes_d.items()}
        self.nes = nes

# Write output file
        if nes_d['Store spectra']:
            nes.storeSpectra(f_out=nes_d['Output file'])


    def los(self):

        logger.info('Starting LOS cone calculation')
        geo = self.get_gui_tab('detector')

        dlos = los.DETECTOR_LOS(geo)

# Plot
        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        fig_los = dlos.plotLOS()
        self.wid.addPlot('Detector LoS', fig_los)
        self.wid.show()

# Write output file
        if geo['Write LOS']:
            dlos.writeLOS()


    def nresp(self):

        nresp_d = self.get_gui_tab('nresp')

        reComp = False
        if not hasattr(self, 'nresp_d'):
            reComp = True
        else:
            for key, val in self.nresp_d.items():
                if key not in ('Energy for PHS plot', 'Write nresp'):
                    if nresp_d[key] != val:
                        reComp = True
                        break
        if reComp:
            nrsp = nresp.NRESP(nresp_d, parallel=nresp_d['MultiProcess'])
        else:
            nrsp = self.nrsp
        fig_nr = nrsp.plotResponse(E_MeV=nresp_d['Energy for PHS plot'])
        if not hasattr(self, 'wid'):
            self.wid = plots.plotWindow()
        self.wid.addPlot('NRESP', fig_nr)
        self.wid.show()
        self.nresp_d = {key: val for key, val in nresp_d.items()}
        self.nrsp = nrsp

# Write output
        if nresp_d['Write nresp']:
            nrsp.to_nresp()


if __name__ == '__main__':


    app = QApplication(sys.argv)
    main = NEREUS()
    app.exec_()



import json
import numpy as np
from scipy.interpolate import interp1d
from .constants import epsilon0, echarge


def sigma_diff(E_in_MeV, mu_in, reac, Z1=None, Z2=None, paired=False):
    '''General method redirecting to the relevant reaction method'''

    if reac in ('DP', 'D3He', 'alphad', 'alphat', 'D(3He,P)4He'):
        return tabulated_sigma_diff(E_in_MeV, mu_in, reac, paired=paired)
    elif reac in ('D(T,n)4He', 'D(D,n)3He'):
        return legendre_sigma_diff(E_in_MeV, mu_in, reac, paired=paired)
    elif Z1 is not None:
        return coulomb_sigma_diff(E_in_MeV, mu_in, Z1, Z2)


def coulomb_sigma_diff(E_in_MeV, mu_in, Z1, Z2):
    '''q1, q2 are the electric charge numbers of the scattering particles'''

    q1 = Z1*echarge
    q2 = Z2*echarge
    E_in_MeV = np.atleast_1d(E_in_MeV)
    mu_in    = np.atleast_1d(mu_in)
    E_in_J = 1e6*echarge*E_in_MeV
    k = 1./(4*np.pi*epsilon0)
    coul = q1*q2*k
    s = coul/np.outer(1 - mu_in, E_in_J) # denominator=0 yields nan, as wished
    cross_sec = 0.25* s**2 * 10**31 # 10**31: m**2 -> mbarn
    return np.squeeze(cross_sec)


def tabulated_sigma_diff(E_in_MeV, mu_in, reac, paired=False):
    '''Performing interp2d on full cross-sectoin table(E, mu)'''

    from scipy.interpolate import interp2d

    f_json = 'crossSections/%s.json' %reac
    with open(f_json, 'r') as fjson:
        cs = json.load(fjson)

    f = interp2d(cs['En'], cs['mu'], cs['sigmaDiff'], kind='linear')

    E_in_MeV = np.atleast_1d(E_in_MeV)
    mu_in    = np.atleast_1d(mu_in)
    if paired:
        res = f(E_in_MeV, mu_in)
    else:
        res = f(E_in_MeV, mu_in).T
        if len(mu_in > 1):
            unsorted = np.argsort(np.argsort(mu_in))
            res = res[..., unsorted]
            res = res.T
    return np.squeeze(res)


def legendre_sigma_diff(E_in_MeV, mu_in, reac, paired=False):

    from scipy.special import eval_legendre

    E_in_MeV = np.atleast_1d(E_in_MeV)
    mu_in    = np.atleast_1d(mu_in)

    f_json = 'crossSections/%s.json' %reac
    with open(f_json, 'r') as fjson:
        cs = json.load(fjson)

    cs['LegCoeff'] = np.array(cs['LegCoeff'])
    n_leg = cs['LegCoeff'].shape[1]
    data = interp1d(cs['En'], cs['LegCoeff'], axis=0)
    data_E = data(E_in_MeV)
    cs_leg = np.zeros((len(mu_in), n_leg))
    for jleg in range(n_leg):
        cs_leg[:, jleg] = eval_legendre(jleg, mu_in)
    if paired:
        cs_sum = np.sum(cs_leg*data_E, axis=1)
    else:
        cs_sum = np.tensordot(cs_leg, data_E, axes=([1, 1]))

    leg_tot = legendre_sigma_tot(E_in_MeV, reac)

    return np.squeeze(leg_tot*cs_sum/(4*np.pi*data_E[:, 0]))


def Poly5(E_keV, poly5):
    return poly5[0] + E_keV*(poly5[1] + E_keV*(poly5[2] + E_keV*(poly5[3] + E_keV*poly5[4])))


def legendre_sigma_tot(E_in_MeV, reac, Emin_MeV=5.e-4):

    f_json = 'crossSections/%s.json' %reac
    with open(f_json, 'r') as fjson:
        cs = json.load(fjson)

    E_in_MeV = np.atleast_1d(E_in_MeV)
    E_keV = 1e3*E_in_MeV
    leg_out = np.zeros_like(E_in_MeV)

    (ind_bosch1, ) = np.where((E_in_MeV >= Emin_MeV) & (E_in_MeV < cs['BoschLiskBound']) & (E_in_MeV <= cs['loHiBound']))
    (ind_bosch2, ) = np.where((E_in_MeV >= Emin_MeV) & (E_in_MeV < cs['BoschLiskBound']) & (E_in_MeV > cs['loHiBound']))
    (ind_interp, ) = np.where((E_in_MeV >= Emin_MeV) & (E_in_MeV >= cs['BoschLiskBound']))
    leg_out[ind_interp] = np.interp(E_in_MeV[ind_interp], cs['Etot_MeV'], cs['y'])

    num_a    = Poly5(E_keV, cs['A'])
    den1_a   = Poly5(E_keV, cs['B'])
    num_ahi  = Poly5(E_keV, cs['Ahi'])
    den1_ahi = Poly5(E_keV, cs['Bhi'])
    den2 = E_keV * np.exp(cs['Bg']/np.sqrt(E_keV))

    leg_out[ind_bosch1] = num_a  [ind_bosch1]/(den1_a  [ind_bosch1]*den2[ind_bosch1])
    leg_out[ind_bosch2] = num_ahi[ind_bosch2]/(den1_ahi[ind_bosch2]*den2[ind_bosch2])
    
    return np.squeeze(leg_out)


if __name__ == '__main__':

# Cross-sections

# Total - correct, page 9 of ControlRoom manual!
    E = 0.5
    dttot = legendre_sigma_tot(E, reac='D(T,n)4He')
    print('DT tot cross-section for E=%8.4f MeV:' %E)
    print('%12.4e millibarn' %dttot)

# Differential - correct, page 9 of ControlRoom manual!
    E = 1
    mu = -1
    print('DT cross-section for E=%8.4f MeV, mu=%6.3f:' %(E, mu))
    dt_diff = sigma_diff(E, mu, reac='D(T,n)4He')
    print('%12.4e millibarn' %dt_diff)

# Differential tab-sigma_diff, DP:
    E = 1
    mu = -1
    print('DP cross-section for E=%8.4f MeV, mu=%6.3f:' %(E, mu))
    dp_diff = sigma_diff(E, mu, reac='dp')
    print('%12.4e millibarn' %dp_diff)

# Differential Coulomnb alpha scattering
    E = 3.5
    mu = -1
    print('Coul cross-section for E=%8.4f MeV, mu=%6.3f:' %(E, mu))
    coul_diff = sigma_diff(E, mu, 'coul', Z1=2, Z2=2)
    print('%12.4e millibarn' %coul_diff)



import numpy as np

def wr_for(arr_in, fmt='%13.6E', n_lin=6):

    arr_flat = arr_in.T.ravel()
    nx = len(arr_flat)
    out_str=''
    for jx in range(nx):
        out_str += (fmt %arr_flat[jx])
        if (jx%n_lin == n_lin - 1):
            out_str += '\n'
# Line break after writing data, but no double break
    if (nx%n_lin != 0):
        out_str += '\n'

    return out_str


def ssplit(ll):

    tmp = ll.replace('-', ' -')
    tmp = tmp.replace('e -', 'e-')
    tmp = tmp.replace('E -', 'E-')
    slist = tmp.split()
    a = [float(i) for i in slist]

    return a


def lines2fltarr(lines, dtyp=None):

    data = []
    for line in lines:
        data += ssplit(line)
    if dtyp is None:
        dtyp = np.float32

    return np.array(data, dtype=dtyp)


def fltarr_len(lines, nx, dtyp=None):

    data = []
    for jlin, line in enumerate(lines):
        data += ssplit(line)
        if len(data) >= nx:
            break
    if dtyp is None:
        dtyp = np.float32

    return jlin+1, np.array(data, dtype=dtyp)



import os, time, logging, json
import numpy as np
from scipy.linalg import norm
from scipy.interpolate import interp1d
import numba as nb

from nresp import crossSections

logger = logging.getLogger('nresp.en2light')
logger.setLevel(level=logging.DEBUG)

nrespDir = os.path.dirname(os.path.realpath(__file__))

# Measure CPU time for several methods

time_cyl  = 0.
count_cyl = 0

PI2 = 2.*np.pi

CS = crossSections.crossSections()
dE = (CS.Egrid[-1] - CS.Egrid[0])/float(len(CS.Egrid) - 1)

flt_typ = np.float64
int_typ = np.int32

mediaCross = np.array([3, 2, 2, 0, 2, 1], dtype=int_typ)

f_mass = '%s/inc/nucleiMassMeV.json' %nrespDir
with open(f_mass, 'r') as fjson:
    massMeV = json.load(fjson)

poly = {"DLT0": 0.0, "DLT1": 0.0, "FLT1": 0.0277, "FLT2": 1.749, "ENAL1": 6.76,
    "GLT0": -0.6366, "GLT1": 0.21, "GLT2": 0.0, "RLT1": 0.01, "SLT1": 0.0097}


@nb.njit
def scatteringDirection(cx_in, ctheta, PHI):

    '''Calculating flight direction of scattered particles in the lab frame
ctheta: cos of scattering angle w.r.t. incident direction
PHI     : azimutal scattering angle 
cx_in : versor of incident particle
cx_out: versor of scattered particle'''

    stheta = np.sqrt(1. - ctheta**2)
    X3 = [stheta*np.cos(PHI), stheta*np.sin(PHI), ctheta]

    if cx_in[2] > 0.999999:
        cx_out = np.array(X3, dtype=flt_typ)
    elif cx_in[2] < -0.999999:
        cx_out = -np.array(X3, dtype=flt_typ)
    else:
        cx_out = cx_in*X3[2]
        S1 = np.sqrt(1. - cx_in[2]**2)
        cx_s = cx_in[:2]/S1
        cx_out[0] +=  cx_s[1]*X3[0] + cx_s[0]*cx_in[2]*X3[1]
        cx_out[1] += -cx_s[0]*X3[0] + cx_s[1]*cx_in[2]*X3[1]
        cx_out[2] += -S1*X3[1]

    return cx_out


def reactionType(ZUU, jEne, reac_list):
    '''Throwing dices for the reaction occurring in a given material'''

    for reac in reac_list:
        ZUU -= CS.cst1d[reac][jEne]
        if ZUU < 0.:
            return reac
    return None


def reactionHC(En_in, alpha_sh, SC, rnd):
    '''Throwing dices for the reaction in a C+H material'''

    ZUU = rnd*(alpha_sh + SC) - alpha_sh
    if ZUU < 0.:
        return 'H(N,N)H'
    jEne = min(int(En_in/dE), len(CS.Egrid)-1)

    return reactionType(ZUU, jEne, ["12C(N,N)12C", "12C(N,N')12C", "12C(N,A)9BE",
        "12C(N,A)9BE'->N+3A", "12C(N,N')3A", "12C(N,P)12B", "12C(N,D)11B"])


def photo_out(elementID, En_in, zr_dl, Egrid, light_int):
    '''Light yield for an arbitrary element'''

    if zr_dl < 0:
        return 0

    if elementID == 1: # H
        if En_in >= Egrid[-1]:
            photo = poly['DLT0'] + poly['DLT1']*En_in
        else:
            photo = light_int(En_in)
    elif elementID == 2: # D
        En = 0.5*En_in
        if En >= Egrid[-1]:
            photo = poly['DLT0'] + poly['DLT1']*En
        else:
            photo = light_int(En)
        photo *= 2.
    elif elementID == 3: # He
        if En_in >= poly['ENAL1']:
            photo = poly['GLT0'] + (poly['GLT1'] + poly['GLT2']*En_in)*En_in
        else:
            photo = poly['FLT1'] * En_in**poly['FLT2']
    elif elementID == 4: # Be
        photo = poly['RLT1']*En_in
    elif elementID == 5: # B, C
        photo = poly['SLT1']*En_in

    return photo


def photo_B8to2alpha(EA1, En_in, CX1, CXS, dEnucl, rnd, Egrid, light_int, mB8_MeV, mHe_MeV):
    '''Light yield of B->2alpha reactions'''

    CTCM = 2.*rnd[0] - 1.
    ctheta, cthetar, enr_loc, ENE = kinema(mB8_MeV, 0., mHe_MeV, dEnucl, CTCM, En_in)
    PHI2 = PI2*rnd[1]
    PHI3 = PHI2 + np.pi
    CX2 = scatteringDirection(CXS, ctheta , PHI2)
    CX3 = scatteringDirection(CXS, cthetar, PHI3)
    CA12 = np.dot(CX1, CX2)
    CA13 = np.dot(CX1, CX3)
    CA23 = np.dot(CX2, CX3)
    elementIndex = [3, 3, 3]
    energy = [EA1, ENE, enr_loc]
    CA0 = 0.999999

    if CA12 >= CA0:
        if energy[0] >= energy[1]:
            elementIndex[1] = 5
        else:
            elementIndex[0] = 5
    if CA13 >= CA0:
        if energy[0] >= energy[2]:
            elementIndex[2] = 5
        else:
            elementIndex[0] = 5
    if CA23 >= CA0:
        if energy[1] >= energy[2]:
            elementIndex[2] = 5
        else:
            elementIndex[1] = 5

    phot_B8to2alpha = 0.
    for j in range(3):
        phot_B8to2alpha += photo_out(elementIndex[j], energy[j], 1., Egrid, light_int)

    return phot_B8to2alpha


@nb.njit
def cylinder_crossing(Radius, H, H0, X0, CX):
    '''Calculating intersections of a straight line crossing a cylinder
z-axis = symmetry axis of the cylinder
If the source point is at the cylinder, this is not counted
as intersection point'''

    H1 = H + H0
    SQ = Radius**2 - X0[0]**2 - X0[1]**2
    C1 = CX[0]**2 + CX[1]**2

    if SQ >= 0 and X0[2] <= H1 and X0[2] >= H0: # Source point inside cylinder, 1 ray intersection
        W1 = 0.
        if C1 < 1E-12: # horizontal
            if CX[2] < 0: W2 = (H0 - X0[2])/CX[2]
            if CX[2] > 0: W2 = (H1 - X0[2])/CX[2]
        else:
            S1 = (CX[0]*X0[0] + CX[1]*X0[1])/C1
            W2 = -S1 + np.sqrt(S1**2 + SQ/C1)
            Z = X0[2] + W2*CX[2]
            if CX[2] < 0:
                if Z < H0: W2 = (H0 - X0[2])/CX[2]
            else:
                if Z > H1: W2 = (H1 - X0[2])/CX[2]
    else:  # Source point outside cylinder
        if C1 <= 0: # horizontal, CX[2] = 1 or -1
            if SQ < 0.:
                return 0., 0.
            W1 = (H0 - X0[2])/CX[2]
            if W1 < 0.:
                return 0., 0.
            W2 = (H1 - X0[2])/CX[2]
            if W2 <= W1:
                W1, W2 = W2, W1
        else:
            S1 = (CX[0]*X0[0] + CX[1]*X0[1])/C1
            SQ2 = S1**2 + SQ/C1
            if SQ2 <= 0.:
                return 0., 0.
            SQ3 = np.sqrt(SQ2)
            W10 = -S1 - SQ3
            W20 = -S1 + SQ3
            Z1 = X0[2] + W10*CX[2]
            Z2 = X0[2] + W20*CX[2]
    
            if CX[2] == 0:
                return 0., 0. # Prevent division by zero
            if Z1 <= H0 and Z2 <= H0:
                return 0., 0.
            if Z1 >= H1 and Z2 >= H1:
                return 0., 0.
# Two ray interesections
            if Z1 > H0 and Z1 < H1:
                Z2 = max(Z2, H0)
                Z2 = min(Z2, H1)
                W1 = W10
                W2 = (Z2 - X0[2])/CX[2]
            elif Z2 > H0 and Z2 < H1:
                Z1 = max(Z1, H0)
                Z1 = min(Z1, H1)
                W2 = W20
                W1 = (Z1 - X0[2])/CX[2]
            else:
                W1 = (H0 - X0[2])/CX[2]
                W2 = (H1 - X0[2])/CX[2]
                if W2 <= W1:
                    W1, W2 = W2, W1
        if W1 == W2 or W2 < 0.:
            return 0., 0.

    return W1, W2


def geom(D, RG, DSZ, RSZ, DL, RL, X0, CX):
    '''Calculating the flight path's crossing points through the three cylinders
MediaSequence    material id:  0 scintillator, 1 light pipe, 2 Al, 3 vacuum (MAT-1)
CrossPathLen     path length to a crossing point(WEG)'''

    global time_cyl, count_cyl

    tcyl1 = time.time()
    W1, W2 = cylinder_crossing(RG , D  , 0., X0, CX) # Outer cylinder
    count_cyl += 1
    time_cyl += time.time() - tcyl1
    if W2 == 0.: # No intersections at all
        return None, None
    tcyl2 = time.time()
    W3, W4 = cylinder_crossing(RSZ, DSZ, DL, X0, CX) # Scintillator
    W5, W6 = cylinder_crossing(RL , DL , 0., X0, CX) # Light guide
    count_cyl += 2
    time_cyl += time.time() - tcyl2

# Mapping paths and medium
    pathl = np.array([W1, W2, W3, W4, W5, W6], dtype=flt_typ)
    IndexPath = PathMedia(pathl)

    return mediaCross[IndexPath], pathl[IndexPath]


@nb.njit
def PathMedia(pathl):

    IndexPath = [j+2 for j, x in enumerate(pathl[2:]) if x]
    if pathl[0] > 0.:
        IndexPath.insert(0, 0)

# Last material is always "3" (vacuum)
    IndexPath.append(1)

# Swap last 2 pairs in case (light guide crossed before scintillator)
    crosspath = pathl[np.array(IndexPath)]
    n_cross = len(crosspath)
    if n_cross > 4:
        if crosspath[-3] <= crosspath[-5]:
            IndexPath[-5:-3], IndexPath[-3:-1] = IndexPath[-3:-1], IndexPath[-5:-3]

# Delete 3rd-last element if its path is longer than the previous
    crosspath = pathl[np.array(IndexPath)]
    if n_cross > 3:
        if crosspath[-3] <= crosspath[-4]:
            del IndexPath[-3]
    if n_cross > 1:
        if crosspath[-1] <= crosspath[-2]:
            del IndexPath[-1]

    return IndexPath


@nb.njit
def kinema(M1, M2, M3, dEnucl, CTCM, T1LAB):
    '''Scattering kinematics (relativistic)'''

    M4 = M1 + M2 - M3 - dEnucl
    m3_sq = M3**2
    E1LAB = T1LAB + M1
    ARG = max(0., E1LAB**2 - M1**2)
    P1LAB = np.sqrt(ARG)
    BETA = P1LAB/(E1LAB + M2)
    GAMMA = 1./np.sqrt(1. - BETA**2)
    ECM = np.sqrt((M1 + M2)**2 + 2.*M2*T1LAB)
    E3CM = (ECM**2 + m3_sq - M4**2)/(2.*ECM)
    ARG = max(0., E3CM**2 - m3_sq)
    P3CM = np.sqrt(ARG)
    E3LAB = GAMMA*(E3CM + BETA*P3CM*CTCM)
    T3LAB = max(0., E3LAB - M3)
    T4LAB = max(0., dEnucl + T1LAB - T3LAB)
    ARG = E3LAB**2 - m3_sq
    if ARG > 0.:
        P3LAB = np.sqrt(ARG)
        ctheta = GAMMA*(P3CM*CTCM + BETA*E3CM)/P3LAB
    else:
        ctheta = 0.
    ARG = T4LAB**2 + 2.*T4LAB*M4 # E4LAB - M4**2
    if ARG > 0.:
        E4CM = ECM - E3CM
        P4CM = P3CM
        P4LAB = np.sqrt(ARG)
        cthetar = GAMMA*(-P4CM*CTCM + BETA*E4CM)/P4LAB
    else:
        cthetar = 0.

    return ctheta, cthetar, T4LAB, T3LAB


def En2light(tuple_in):

    En_in_MeV, phs_max, nresp_set = tuple_in
    Egrid, det_light = np.loadtxt(nresp_set['f_in_light'], skiprows=1, unpack=True)
    light_int = interp1d(Egrid, det_light, assume_sorted=True, fill_value='extrapolate')

    with open(nresp_set['f_detector'], 'r') as fjson:
        detector = json.load(fjson)

    nmc = int(nresp_set['nmc'])
    En_wid = nresp_set['En_wid_frac']*En_in_MeV

    time_reac = {}
    for reac in CS.reacTotUse:
        time_reac[reac] = 0.
    tim = np.zeros(6)
    time_slow = np.zeros(len(tim) - 1)
    time_pre = 0.

    GWT_EXP = np.zeros(6, dtype=flt_typ)
    SIGM = np.zeros(4, dtype=flt_typ)
    X00  = np.zeros(3, dtype=flt_typ)
    X0   = np.zeros_like(X00)
    CX   = np.zeros_like(X00)

# Derived quantities

    rsz_sq = detector['RSZ']**2
    rg_sq  = detector['RG' ]**2
    detector['D']  = detector['DG'] + detector['DL']
    detector['RL'] = detector['RSZ']
    XNC  = detector['dens_sc']*6.023*massMeV['amu']/(detector['alpha_sc']*massMeV['H'] + massMeV['C12'])
    XNH  = detector['alpha_sc']*XNC
    XNCL = detector['dens_lg']*6.023*massMeV['amu']/(detector['alpha_lg']*massMeV['H'] + massMeV['C12'])
    XNHL = detector['alpha_lg']*XNCL
    XNAL = 0.60316

    the = np.radians(detector['theta'])
    cos_the = np.cos(the)
    sin_the = np.sin(the)
    if sin_the == 0.:
        cotan_the = 0. #0.5*np.pi
    else:
        cotan_the = cos_the/sin_the

# Geometry

    if detector['theta'] <= 0. :
        CTMAX = 0.
        distance = detector['dist'] - (detector['DG'] - 0.5*detector['DSZ'])
        if distance > 0.:
            CTMAX = 1./np.sqrt(1. + (detector['RG']/distance)**2)
    elif detector['theta'] < 90. :
        if sin_the < 0.999:
            R0 = detector['RG'] + detector['D']*sin_the/cos_the
    else:
        RR  = detector['RG']*np.sqrt(1. - (detector['RG']/detector['dist'])**2)
        ASS = detector['RG']/RR*(detector['dist'] - np.sqrt(rg_sq - RR**2))

    X00[0] = detector['dist']*sin_the
    X00[1] = 0.
    X00[2] = detector['DL'] + 0.5*detector['DSZ'] + detector['dist']*cos_the

    logger.info('START - Eneut: %8.4f MeV', En_in_MeV)

# Intialise

    n_react = len(CS.reacTotUse) + 1
    count_reac   = np.zeros(n_react     , dtype=int_typ)
    count_pp3as  = np.zeros(CS.max_level, dtype=int_typ)
    phs_dim_rea  = np.zeros(n_react     , dtype=flt_typ)
    phs_dim_pp3  = np.zeros(CS.max_level, dtype=flt_typ)
    light_output = np.zeros((n_react     , phs_max), dtype=flt_typ)
    pp3as_output = np.zeros((CS.max_level, phs_max), dtype=flt_typ)

    n_rand   = min(300000, nmc*10)
    n_rand1  = int(0.99*n_rand)
    ng_rand  = min(300000, nmc)
    ng_rand1 = int(0.99*ng_rand)
    jrand  = n_rand
    jg_rnd = ng_rand

# MC loop
    np.random.seed(0)
    for j_mc in range(nmc):
        tbeg = time.time()
        if jrand > n_rand1:
            jrand = 0
            rand = np.random.random_sample(n_rand)
            logger.debug('Re-random flat %d %d', j_mc, nmc)
        weight = 1.

        if detector['theta'] == 0.: # Source position at 0 deg
            if CTMAX > 0.9999:
                RR0 = detector['RG']*np.sqrt(rand[jrand])
                FI0 = PI2*rand[jrand+1]
                jrand += 2
                X0[0] = RR0*np.cos(FI0)
                X0[1] = RR0*np.sin(FI0)
                X0[2] = detector['D']
                CX[0] = 0.
                CX[1] = 0.
                CX[2] = -1.
            else:
                X0[0] = 1e6
                while np.abs(X0[0]) >= detector['RG']:
                    CX[2] = -1. + (1. - CTMAX)*rand[jrand]
                    jrand += 1
                    if CX[2] <= -1.:
                        CX[2] = -1. + 1E-10
                    CX[0] = np.sqrt(1. - CX[2]**2)
                    X0[0] = -distance*CX[0]/CX[2]
                CX[1] = 0.
                X0[1] = 0.
                X0[2] = detector['D']
        elif detector['theta'] == 90.: # Source position at 90 deg, versor perp to the scintillator axis
            X0[2] = 0.
            while (X0[2] == detector['D'] or X0[2] == 0.):
                X0[1] = RR*(2*rand[jrand] - 1.)
                X0[2] = detector['D']*rand[jrand+1]
                jrand += 2
            X0[0] = np.sqrt(rg_sq - X0[1]**2)
            H1 = norm(X00 - X0)
            CX = (X0 - X00)/H1
        else:  # (only for distance > 500 cm, assuming parallel neutron beam)
            CX[0] = -sin_the
            CX[1] = 0.
            CX[2] = -cos_the
            X0[1] = detector['RG']*(2.*rand[jrand] - 1.)
            H1 = np.sqrt(rg_sq - X0[1]**2)
            X0[0] = -detector['RG'] + (R0 + detector['RG'])*rand[jrand+1]
            jrand += 2
            while (-X0[0] >= H1 or X0[0] >= H1 + R0 - detector['RG']):
                X0[0] = -detector['RG'] + (R0 + detector['RG'])*rand[jrand]
                jrand += 1
            X0[2] = detector['D']
            if X0[0] > H1:
                X0[2] = detector['D'] - (X0[0] - H1)*cotan_the
                X0[0] = H1

        n_scat = 0
        LightYieldChain = 0.
        LEVEL0 = 0
        ENE = En_in_MeV # MeV

        if nresp_set['distr'] == 'gauss':
            if jg_rnd > ng_rand1:
                logger.debug('Re-random gauss')
                jg_rnd = 0
                gauss_rnd = np.random.normal(loc=En_in_MeV, scale=En_wid, size=ng_rand)
            ENE = gauss_rnd[jg_rnd]
            jg_rnd += 1

        tend = time.time()
        time_pre += tend - tbeg

# Chain of reactions
        while(True):

# Flight path
# Input: D, RG, DSZ, RSZ, DL, RL, X0[2], CX[2]
            tim[0] = time.time()
            MediaSequence, CrossPathLen = geom(detector['D'], detector['RG'], detector['DSZ'], detector['RSZ'], detector['DL'], detector['RL'], X0, CX)
            tim[1] = time.time()

            if MediaSequence is None:
                break

            tim[2] = time.time()
            n_cross_cyl = len(MediaSequence)
            jEne = min(int(ENE/dE), len(CS.Egrid)-1)
            SH  = CS.cst1d['H(N,N)H'][jEne] # No log, different from fortran
            SC  = CS.cst1d['CarTot'][jEne]
            SAL = CS.cst1d['AlTot' ][jEne]
            SIGM[0] = XNH*SH  + XNC*SC
            SIGM[1] = XNHL*SH + XNCL*SC
            SIGM[2] = XNAL*SAL

            SIG = 1e-4*SIGM[MediaSequence]
            RHO = rand[jrand]
            jrand += 1

            GWT_EXP[0] = CrossPathLen[0]*SIG[0]
            for I in range(1, n_cross_cyl):
                GWT_EXP[I] = GWT_EXP[I-1] + (CrossPathLen[I] - CrossPathLen[I-1])*SIG[I]

            RGWT = 1. - np.exp(-GWT_EXP)

            if n_scat <= 0:
                weight = RGWT[n_cross_cyl-1]
                RHO *= weight

            log_RHO = np.log(1. - RHO)
            MediumID = 3
            PathInMedium = 0.
            if SIG[0] > 0. and RGWT[0] >= RHO:
                PathInMedium = -log_RHO/SIG[0]
                MediumID = MediaSequence[0]
            else:
                for I in range(1, n_cross_cyl):
                    if RGWT[I] >= RHO:
                        PathInMedium = CrossPathLen[I-1] - (GWT_EXP[I-1] + log_RHO)/SIG[I]
                        MediumID = MediaSequence[I]
                        break

            if PathInMedium > CrossPathLen[n_cross_cyl-1]:
                MediumID = 3

            XR = X0 + PathInMedium*CX
            zr_dl = XR[2] - detector['DL']
            tim[3] = time.time()

            if weight < 2.E-5 or MediumID == 3:
                break # Reac. chain
            tim[4] = time.time()
            n_scat += 1

# Random scattering angle

            PHI = PI2*rand[jrand]
            Frnd = rand[jrand+1]
            jrand += 2

            tim[5] = time.time()
            time_slow += np.diff(tim)

            if MediumID in (0, 1):

#---------------------
# Random reaction type
#---------------------

                reac_type = None
                if MediumID == 0:
                    alpha_sh = detector['alpha_sc']*SH #XNH/XNC
                else:
                    alpha_sh = detector['alpha_lg']*SH #XNHL/XNCL
                while reac_type is None:
                    reac_type = reactionHC(ENE, alpha_sh, SC, rand[jrand])
                    jrand += 1

                if n_scat == 1: # Label first neutron reaction
                    if MediumID == 0:
                        first_reac_type = CS.reacTotUse.index(reac_type)
                    elif MediumID == 1:
                        first_reac_type = n_react - 1 # 1st reaction (no matter which) in light guide

#-----------
# Kinematics
#-----------

                if reac_type == 'H(N,N)H':
                    tbeg = time.time()
                    CTCM = 2.*Frnd - 1.
                    if ENE > 2.:   # Angular distribution
                        AAA = CS.cst1d['HE1'][jEne]
                        BBB = CS.cst1d['HE2'][jEne]
                        CTCM1 = (CTCM + AAA)/(1. - BBB + AAA*CTCM  + BBB*CTCM **2)
                        CTCM  = (CTCM + AAA)/(1. - BBB + AAA*CTCM1 + BBB*CTCM1**2)
                    dEnucl = CS.crSec_d[reac_type]['dEnucl']
                    ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['H'], massMeV['neutron'], dEnucl, CTCM, ENE)
                    if zr_dl >= 0.:
                        if ENR <= 0.2: 
                            BR = 1.507E-3*ENR
                        else:
                            BR = 2.0457E-3*(ENR + 0.15045)**1.8194
                        LightYieldChain += photo_out(1, ENR, zr_dl, Egrid, light_int)
                        rsz_sq_xr = rsz_sq - XR[0]**2 - XR[1]**2
                        if zr_dl <= BR or detector['DSZ'] - zr_dl <= BR or rsz_sq_xr <= 2.*detector['RSZ']*BR:
                            PHIR = PHI - np.pi
                            CXR = scatteringDirection(CX, cthetar, PHIR)
                            if CXR[2] < 0:
                                WMZ = -zr_dl/CXR[2]
                            elif CXR[2] > 0:
                                WMZ = (detector['DSZ'] - zr_dl)/CXR[2]
                            if np.abs(CXR[2]) > 0.9999:
                                PATHM = WMZ
                            else:
                                cxr_fac = 1./(1. - CXR[2]**2)
                                WR = (XR[0]*CXR[0] + XR[1]*CXR[1])  * cxr_fac
                                WM = rsz_sq_xr * cxr_fac
                                if WM <= 0.:
                                    WMR = -WR + np.abs(WR)
                                else:
                                    WMR = -WR + np.sqrt(WM + WR**2)
                                if WMR < WMZ or CXR[2] == 0.:
                                    PATHM = WMR
                                else:
                                    PATHM = WMZ
                            PATH = BR - PATHM
                            if PATH > 0.:
                                if PATH > 3.104E-4:
                                    ENT = -0.150 + (PATH*488.83)**0.5496
                                else:
                                    ENT = 663.57*PATH
                                LightYieldChain -= photo_out(1, ENT, zr_dl, Egrid, light_int)
                    time_reac[reac_type] += time.time() - tbeg

                elif reac_type in ('12C(N,N)12C', "12C(N,N')12C"):
                    tbeg = time.time()
                    CTCM = CS.cosInterpReac2d(reac_type, ENE, Frnd) # elastic
                    dEnucl = CS.crSec_d[reac_type]['dEnucl']
                    ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['C12'], massMeV['neutron'], dEnucl, CTCM, ENE)
                    LightYieldChain += photo_out(5, ENR, zr_dl, Egrid, light_int)
                    time_reac[reac_type] += time.time() - tbeg

                elif reac_type == '12C(N,A)9BE':
                    tbeg = time.time()
                    if zr_dl > 0.:
                        CTCM = CS.cosInterpReac2d(reac_type, ENE, Frnd)
                        dEnucl = CS.crSec_d[reac_type]['dEnucl']
                        ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['C12'], massMeV['He'], dEnucl, CTCM, ENE)
                        LightYieldChain += photo_out(3, ENE, zr_dl, Egrid, light_int) + photo_out(4, ENR, zr_dl, Egrid, light_int)
                    time_reac[reac_type] += time.time() - tbeg
                    break # reac. chain

                elif reac_type =="12C(N,A)9BE'->N+3A":
                    tbeg = time.time()
                    CTCM = 2.*Frnd - 1.
                    dEnucl = CS.crSec_d[reac_type]['dEnucl']
                    ctheta, cthetar, ENR, EA1 = kinema(massMeV['neutron'], massMeV['C12'], massMeV['He'], dEnucl, CTCM, ENE)
                    CX1 = scatteringDirection(CX, ctheta, PHI)
                    PHI += np.pi
                    CX = scatteringDirection(CX, cthetar, PHI)
                    CTCM = 2.*rand[jrand] - 1.
                    jrand += 1
                    dEnucl = 0.761
                    ctheta, cthetar, ENR, ENE = kinema(massMeV['B9'], 0., massMeV['neutron'], dEnucl, CTCM, ENR)
                    if zr_dl >= 0.:
                        PHI = PI2*rand[jrand] + np.pi
                        jrand += 1
                        CXS = scatteringDirection(CX, cthetar, PHI)
                        dEnucl = 0.095
                        if n_scat == 1:
                            LEVEL0 = 10
                        LightYieldChain += photo_B8to2alpha(EA1, ENR, CX1, CXS, dEnucl, rand[jrand:jrand+2], Egrid, light_int, massMeV['B8'], massMeV['He'])
                        jrand += 2
                    time_reac[reac_type] += time.time() - tbeg

                elif reac_type == "12C(N,N')3A":
                    tbeg = time.time()
                    LEVEL = 0
                    if ENE >= 10.:
                        NRA = rand[jrand]
                        jrand += 1
                        NL = 0.
                        tmp = CS.int_alphas3(ENE)
                        for LEVEL in range(CS.max_level):
                            NL += tmp[LEVEL]
                            if NL >= NRA:
                                break
                    dEnucl = CS.alphas3['q3a'][LEVEL]
                    if LEVEL == 1:
                        CTCM = CS.cosInterpReac2d("12C(N,N')12C*", ENE, Frnd)
                    else:
                        CTCM = 2.*Frnd - 1.
                    ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['C12'], massMeV['neutron'], dEnucl, CTCM, ENE)
                    if zr_dl >= 0.:
                        PHI += np.pi
                        CXR = scatteringDirection(CX, cthetar, PHI)
                        CTCM = 2.*rand[jrand] - 1.
                        PHI1 = PI2*rand[jrand+1]
                        jrand += 2
                        if n_scat == 1:
                            LEVEL0 = LEVEL
                        dEnucl = -dEnucl - 7.369
                        if LEVEL > 1:
                            LEX = rand[jrand]
                            jrand += 1
                            if LEX <= CS.alphas3['3MeV'][LEVEL]:
                                dEnucl -= 3.
                        ctheta1, cthetar, ENR, EA1 = kinema(massMeV['C12'], 0., massMeV['He'], dEnucl, CTCM, ENR)
                        CX1 = scatteringDirection(CXR, ctheta1, PHI1)
                        PHIR = PHI1 + np.pi
                        CXS = scatteringDirection(CXR, cthetar, PHIR)
                        dEnucl = 0.095
                        if LEVEL > 1 and LEX <= CS.alphas3['3MeV'][LEVEL]:
                            dEnucl += 3.
                        LightYieldChain += photo_B8to2alpha(EA1, ENR, CX1, CXS, dEnucl, rand[jrand: jrand+2], Egrid, light_int, massMeV['B8'], massMeV['He'])
                        jrand += 2
                    time_reac[reac_type] += time.time() - tbeg

                elif reac_type == '12C(N,P)12B':
                    tbeg = time.time()
                    if zr_dl > 0.:
                        CTCM = 2.*Frnd - 1.
                        dEnucl = CS.crSec_d[reac_type]['dEnucl']
                        ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['C12'], massMeV['H'], dEnucl, CTCM, ENE)
                        LightYieldChain += photo_out(1, ENE, zr_dl, Egrid, light_int) + photo_out(5, ENR, zr_dl, Egrid, light_int)
                    time_reac[reac_type] += time.time() - tbeg
                    break # reac_chain

                elif reac_type == '12C(N,D)11B':
                    tbeg = time.time()
                    if zr_dl > 0.:
                        CTCM = 2.*Frnd - 1.
                        dEnucl = CS.crSec_d[reac_type]['dEnucl']
                        ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['C12'], massMeV['D'], dEnucl, CTCM, ENE)
                        LightYieldChain += photo_out(2, ENE, zr_dl, Egrid, light_int) + photo_out(5, ENR, zr_dl, Egrid, light_int)
                    time_reac[reac_type] += time.time() - tbeg
                    break # reac_chain

# Reaction in aluminium cage
            elif MediumID == 2:
                ZUU = rand[jrand]*SAL
                jrand += 1
                reac_type = reactionType(ZUU, jEne, ["27AL(N,N)27AL", "27AL(N,N')27AL'"])
                if reac_type is None:
                    break #reac_chain
                if n_scat <= 1:
                    first_reac_type = CS.reacTotUse.index(reac_type)

                if reac_type == '27AL(N,N)27AL':
                    CTCM = CS.cosInterpReac2d(reac_type, ENE, Frnd)
                    dEnucl = CS.crSec_d[reac_type]['dEnucl']
                elif reac_type == "27Al(N, N')27Al":
                    CTCM = 2.*Frnd - 1.
                    dEnucl = -rand[jrand]*(ENE - 0.5) - 0.5
                    jrand += 1
                ctheta, cthetar, ENR, ENE = kinema(massMeV['neutron'], massMeV['Al'], massMeV['neutron'], dEnucl, CTCM, ENE)

            if ENE <= 0.01:
                break #reac_chain
            CX = scatteringDirection(CX, ctheta, PHI)
            X0 = XR

# End reaction chain

        if weight >= 2E-5 and LightYieldChain > 0.:
            phsBin = int(LightYieldChain/nresp_set['Ebin_MeVee'])
            light_output[first_reac_type, phsBin] += weight
            count_reac  [first_reac_type] += 1 # count
            phs_dim_rea [first_reac_type] = max(phsBin, phs_dim_rea[first_reac_type])
            if first_reac_type == 5:
                count_pp3as [LEVEL0] += 1
                phs_dim_pp3 [LEVEL0] = max(phsBin, phs_dim_pp3[LEVEL0])
                pp3as_output[LEVEL0, phsBin] += weight

# End Energy MC loop (incoming neutron)

# Unnormalise arrays w.r.t. NMC and viewing solid angle

    norm_mc_F0 = (np.pi*rg_sq*cos_the + 2.*detector['D']*detector['RG']*sin_the)/float(nmc*nresp_set['Ebin_MeVee'])
    light_output *= norm_mc_F0
    pp3as_output *= norm_mc_F0

    logger.debug('\nTime analysis:')
    logger.debug('cyl_cross %8.4f %d', time_cyl , count_cyl)
    logger.debug('time_pre %8.4f', time_pre)
    for reac in CS.reacTotUse:
        logger.debug('%s %8.4f', reac, time_reac[reac])
    logger.debug('Bottle necks in energy chain, En=%8.4f MeV' %En_in_MeV)
    for ts in time_slow:
        logger.debug(ts)

    return count_reac, count_pp3as, phs_dim_rea, phs_dim_pp3, light_output, pp3as_output



from . import constants as con

class reac_class:
    pass

class proton:
    m = con.mpc2
    Z = 1

class neutron:
    m = con.mnc2
    Z = 0

class deuteron:
    m = con.mDc2
    Z = 1

class triton:
    m = con.mTc2
    Z = 1

class he3:
    m = con.mHe3c2
    Z = 2

class he4:
    m = con.mHe4c2
    Z = 2

# Reaction classes
# neutrons/paper/bosch_nf92.pdf page 625 table VII

reaction = {}

reaction['DP'] = reac_class()
reaction['DP'].in1 = proton
reaction['DP'].in2 = deuteron

reaction['D(D,n)3He'] = reac_class()
reaction['D(D,n)3He'].in1 = deuteron
reaction['D(D,n)3He'].in2 = deuteron
reaction['D(D,n)3He'].prod1 = neutron
reaction['D(D,n)3He'].prod2 = he3
reaction['D(D,n)3He'].coeff_reac = [5.4336e-12, 5.85778e-3, 7.68222-3, 0, -2.964e-6, 0, 0]

reaction['D(D,P)T'] = reac_class()
reaction['D(D,P)T'].in1 = deuteron
reaction['D(D,P)T'].in2 = deuteron
reaction['D(D,P)T'].prod1 = proton
reaction['D(D,P)T'].prod2 = triton
reaction['D(D,P)T'].coeff_reac = [5.65718e-12, 3.41267e-3, 1.99167e-3, 0, 1.05060e-5, 0, 0]

reaction['D(3He,P)4He'] = reac_class()
reaction['D(3He,P)4He'].in1 = he3
reaction['D(3He,P)4He'].in2 = deuteron
reaction['D(3He,P)4He'].prod1 = proton
reaction['D(3He,P)4He'].prod2 = he4
reaction['D(3He,P)4He'].coeff_reac = [5.51036e-10, 6.41918e-3, -2.02896e-3, -1.9108e-5,  1.35776e-4, 0, 0]

reaction['D(T,n)4He'] = reac_class()
reaction['D(T,n)4He'].in1 = deuteron
reaction['D(T,n)4He'].in2 = triton
reaction['D(T,n)4He'].prod1 = neutron
reaction['D(T,n)4He'].prod2 = he4
reaction['D(T,n)4He'].coeff_reac = [1.17302e-9 , 1.51361e-2, 7.51886e-2, 4.60643e-3,  1.35000e-2, -1.06750e-4, 1.36600e-5]
